"""
run_general
RD
Jan 22 2020

This is the starting point to run the general system. This general system (formerly optics_propagate.py) is fully
customisable from this script from a series of toggles.

"""

import numpy as np

from medis.params import sp, ap, tp
from medis.utils import dprint
from medis.plot_tools import view_spectra, view_timeseries, quick2D, plot_planes, grid
import medis.medis_main as mm



#################################################################################################
#################################################################################################
#################################################################################################

tp.prescription = 'general_telescope'

# Companion
ap.companion = True
ap.contrast = [1e-5]
ap.companion_xy = [[15, -15]]  # units of this are in lambda/tp.entrance_d

sp.numframes = 2
sp.focused_sys = False
sp.beam_ratio = 0.3  # parameter dealing with the sampling of the beam in the pupil/focal plane
sp.grid_size = 512  # creates a nxn array of samples of the wavefront
sp.maskd_size = 512  # will truncate grid_size to this range (avoids FFT artifacts) # set to grid_size if undesired
sp.closed_loop = False

# Toggles for Aberrations and Control
tp.entrance_d = 8
tp.obscure = False
tp.use_atmos = True
tp.use_ao = True
tp.ao_act = 60
tp.rotate_atmos = False
tp.rotate_sky = False
tp.f_lens = 200.0 * tp.entrance_d
tp.occult_loc = [0,0]

# Saving
sp.save_to_disk = False  # save obs_sequence (timestep, wavelength, x, y)
sp.save_list = ['detector']  # list of locations in optics train to save
# sp.skip_planes = ['coronagraph']  # ['wfs', 'deformable mirror']  # list of locations in optics train to save
sp.quick_detect = True
sp.debug = False
sp.verbose = True

if __name__ == '__main__':
    # =======================================================================
    # Run it!!!!!!!!!!!!!!!!!
    # =======================================================================

    sim = mm.RunMedis(name='general_example', product='photons')
    observation = sim()
    fp_sampling = sim.camera.platescale
    rebinned_photons = sim.camera.rebinned_cube

    dprint(f"Sampling in focal plane is {fp_sampling})
    for o in range(len(ap.contrast)+1):
        print(rebinned_photons.shape)
        datacube = rebinned_photons[o]
        print(o, datacube.shape)
        grid(datacube, logZ=True, title='Spectral Channels')



'''All credit for this function goes to Eelco Hoogendoom at
stackoverflow.com/questions/21100716/fast-arbitrary-distribution-random-sampling'''

import numpy as np
import scipy.misc
import scipy.special as special

class Distribution(object):
    """
    draws samples from a one dimensional probability distribution,
    by means of inversion of a discrete inverstion of a cumulative density function

    the pdf can be sorted first to prevent numerical error in the cumulative sum
    this is set as default; for big density functions with high contrast,
    it is absolutely necessary, and for small density functions,
    the overhead is minimal

    a call to this distibution object returns indices into density array
    """
    def __init__(self, pdf, sort = True, interpolation = True, transform = lambda x: x):
        self.shape          = pdf.shape
        self.pdf            = pdf.ravel()
        self.sort           = sort
        self.interpolation  = interpolation
        self.transform      = transform

        # a pdf can not be negative
        assert(np.all(pdf>=0))

        # sort the pdf by magnitude
        if self.sort:
            self.sortindex = np.argsort(self.pdf, axis=None)
            self.pdf = self.pdf[self.sortindex]
        # construct the cumulative distribution function
        self.cdf = np.cumsum(self.pdf)

    @property
    def ndim(self):
        return len(self.shape)

    @property
    def sum(self):
        """cached sum of all pdf values; the pdf need not sum to one, and is imlpicitly normalized"""
        return self.cdf[-1]

    def __call__(self, N):
        """draw """
        # pick numbers which are uniformly random over the cumulative distribution function
        # print N, self.ndim, self.sum
        choice = np.random.uniform(high = self.sum, size = N)
        # find the indices corresponding to this point on the CDF
        index = np.searchsorted(self.cdf, choice)
        # if necessary, map the indices back to their original ordering
        if self.sort:
            index = self.sortindex[index]
        # map back to multi-dimensional indexing
        index = np.unravel_index(index, self.shape)
        index = np.vstack(index)
        # is this a discrete or piecewise continuous distribution?
        if self.interpolation:
            index = np.float_(index)
            # index[0] += np.random.uniform(size=index.shape[1])
            index[:int(self.interpolation)] += np.random.uniform(size=(int(self.interpolation), index.shape[1]))

        return self.transform(index)


def planck(T, l):
    from scipy.constants import codata
    import numpy as np
    import matplotlib.pyplot as plt

    D = codata.physical_constants

    h = D['Planck constant'][0]
    k = D['Boltzmann constant'][0]
    c = D['speed of light in vacuum'][0]
    wienConstant = 2.897e-3
    # calculate the Planck Law for a specific temperature and an array of wavelengths
    p = c*h/(k*l*T)
    result = np.zeros(np.shape(l))+1e-99
    # prevent underflow - compute only when p is "not too big"
    calcMe = np.where(p<700)
    result[calcMe] = (h*c*c)/(np.power(l[calcMe], 5.0) * (np.exp(p[calcMe])-1))
    return result


def poisson(lamda, k):
    pdf = (lamda ** k * np.exp(-lamda)) / scipy.misc.factorial(k)
    return pdf


def bessel(k):
    '''modified zero order besset=l'''
    pdf = scipy.special.i0(k)
    return pdf


def MR(I,Ic,Is):
    '''modified rician distribution'''
    # print np.exp(-1.0*(I+Ic)/Is)[::10], special.iv(0,2.0*np.sqrt(I*Ic)/Is)[::10] 
    # print Ic, Is
    pdf = 1.0/Is * np.exp(-1.0*(I+Ic)/Is)* special.iv(0,2.0*np.sqrt(I*Ic)/Is)
    #pdf = (1./Is) * np.exp(-(I+Ic)/Is) * bessel(2*np.sqrt(I*Ic)/Is)
    pdf = pdf/sum(pdf)
    return pdf


def gaussian(mu, sig, x):
    pdf = np.exp(-np.power(x - mu, 2.) / (2 * np.power(sig, 2.)))
    return pdf


def gaussian2(x, sig,mu):
    pdf = np.exp(-np.power(x - mu, 2.) / (2 * np.power(sig, 2.)))
    pdf = pdf/sum(pdf)
    return pdf


def lognorm(x,mu, sigma):
    pdf = (np.exp(-(np.log10(x - mu))**2 / (2 * sigma**2))/ (x * sigma))
    pdf = pdf/sum(pdf)
    return pdf


if __name__=='__main__':
    res_elements = 1000
    I = np.linspace(0.1, 200, res_elements)

    ratios = [0,0.5,1,2,7]
    for ratio in ratios:
        Is = 10
        pdf = MR(Is*ratio,Is,I)
        pdf = pdf/ np.sum(pdf)
        # plt.plot(pdf)


        dist = Distribution(pdf)#, transform=lambda i: i - res_elements/2)
        TOAs = dist(100)
        print(TOAs, np.shape(TOAs))
        hist, bins = np.histogram(TOAs, bins ='auto')
        plt.plot(bins[:-1], hist)

    # plt.plot(dist(1000),'o')
    plt.show()

    # shape = 3,3
    # pdf = np.ones(shape)
    # pdf[1]=0
    # dist = Distribution(pdf, transform=lambda i:i-1.5)
    # print dist(10)
    # import matplotlib.pyplot as pp
    # pp.scatter(*dist(1000))
    # pp.show()


"""
Prescription containing a full system that can be fully toggled using switches in the parent module

*** This module is currently unverified ***

TODO
    * verify this script using all the modes
    * reimplement tiptilt again?

"""

import proper

import medis.atmosphere as atmos
import medis.adaptive as ao
import medis.aberrations as aber
import medis.optics as opx
from medis.coronagraphy import coronagraph
from medis.params import iop, sp, tp, atmp

# todo convert parameterss defined here to their own prescription params class
tp.lens_params = [{'aber_vals': [5e-18, 2.0, 3.1], 'diam': 0.2,  'focal_length': 1.2, 'dist' : 1.345, 'name': 'CPA'},
                  {'aber_vals': [5e-18, 2.0, 3.1], 'diam': 0.2, 'focal_length': 1.2, 'dist': 1.345, 'name': 'NCPA'}]

def general_telescope(empty_lamda, grid_size, PASSVALUE):
    """
    #TODO pass complex datacube for photon phases

    propagates instantaneous complex E-field through the optical system in loop over wavelength range

    this function is called as a 'prescription' by proper

    uses PyPROPER3 to generate the complex E-field at the source, then propagates it through atmosphere, then telescope, to the focal plane
    currently: optics system "hard coded" as single aperture and lens
    the AO simulator happens here
    this does not include the observation of the wavefront by the detector
    :returns spectral cube at instantaneous time
    """
    # print("Propagating Broadband Wavefront Through Telescope")

    wfo = opx.Wavefronts(debug=sp.debug)
    wfo.initialize_proper(set_up_beam=True)

    ###################################################
    # Atmosphere, and Secondary Obscuration
    ###################################################

    # Pass through a mini-atmosphere inside the telescope baffle
    #  The atmospheric model used here (as of 3/5/19) uses different scale heights,
    #  wind speeds, etc to generate an atmosphere, but then flattens it all into
    #  a single phase mask. The phase mask is a real-valued delay lengths across
    #  the array from infinity. The delay length thus corresponds to a different
    #  phase offset at a particular frequency.
    if tp.use_atmos:
        wfo.loop_collection(atmos.add_atmos, PASSVALUE['iter'],
                            (iop.atmosdir, sp.sample_time, atmp.model),
                            spatial_zoom=True, plane_name='atmosphere')

    #TODO rotate atmos not yet implementid in 2.0
    # if tp.rotate_atmos:
    #     wfo.loop_collection(aber.rotate_atmos, *(PASSVALUE['iter']))

    # Both offsets and scales the companion wavefront
    if wfo.wf_collection.shape[1] > 1:
        wfo.loop_collection(opx.offset_companion)
        wfo.loop_collection(proper.prop_circular_aperture,
                            **{'radius': tp.entrance_d / 2})  # clear inside, dark outside

    # TODO rotate atmos not yet implementid in 2.0
    # if tp.rotate_sky:
    #     wfo.loop_collection(opx.rotate_sky, *PASSVALUE['iter'])

    ########################################
    # Telescope Primary-ish Aberrations
    #######################################
    # Abberations before AO

    wfo.loop_collection(aber.add_aber, iop.aberdir, PASSVALUE['iter'], lens_name='CPA')

    # wfo.loop_collection(proper.prop_circular_aperture, **{'radius': tp.entrance_d / 2})
    # wfo.wf_collection = aber.abs_zeros(wfo.wf_collection)

    #######################################
    # AO
    #######################################
    if tp.use_ao:

        if sp.closed_loop:
            previous_output = ao.retro_wfs(PASSVALUE['AO_field'], wfo, plane_name='wfs')  # unwrap a previous steps phase map
            wfo.loop_collection(ao.deformable_mirror, WFS_map=None, iter=PASSVALUE['iter'],
                                previous_output=previous_output, plane_name='deformable mirror', zero_outside=True)
        elif sp.ao_delay > 0:
            WFS_map = ao.retro_wfs(PASSVALUE['WFS_field'], wfo, plane_name='wfs')  # unwrap a previous steps phase map
            wfo.loop_collection(ao.deformable_mirror, WFS_map, iter=PASSVALUE['iter'], previous_output=None,
                                plane_name='deformable mirror', zero_outside=True)
        else:
            WFS_map = ao.open_loop_wfs(wfo, plane_name='wfs')  # just uwraps this steps measured phase_map
            wfo.loop_collection(ao.deformable_mirror, WFS_map, iter=PASSVALUE['iter'], previous_output=None,
                                plane_name='deformable mirror', zero_outside=True)

    # Obscure Baffle
    if tp.obscure:
        wfo.loop_collection(opx.add_obscurations, M2_frac=1/8, d_primary=tp.entrance_d,
                            legs_frac=tp.legs_frac)

   ########################################
    # Post-AO Telescope Distortions
    # #######################################
    # Abberations after the AO Loop

    wfo.loop_collection(aber.add_aber, iop.aberdir, PASSVALUE['iter'], lens_name='NCPA')
    wfo.loop_collection(proper.prop_circular_aperture, **{'radius': tp.entrance_d / 2})
    # TODO does this need to be here?
    # wfo.loop_collection(opx.add_obscurations, tp.entrance_d/4, legs=False)
    # wfo.wf_collection = aber.abs_zeros(wfo.wf_collection)

    wfo.loop_collection(opx.prop_pass_lens, tp.lens_params[0]['focal_length'],
                        tp.lens_params[0]['focal_length'], plane_name='pre_coron')


    ########################################
    # Coronagraph
    ########################################
    # there are additional un-aberated optics in the coronagraph module

    wfo.loop_collection(coronagraph, occulter_mode=tp.cg_type, plane_name='coronagraph')

    ########################################
    # Focal Plane
    ########################################
    cpx_planes, sampling = wfo.focal_plane()

    print(f"Finished datacube at timestep = {PASSVALUE['iter']}")

    return cpx_planes, wfo.plane_sampling



"""
adaptive.py

functions relating to simulating an AO system with Proper
mostly code copied from the original MEDIS from Rupert

Generally, the optical prescription will call the deformable_mirror function, which will compile all information and
run sequentially all functions related to creating the adaptive optic correction (main AO functionality, relating to
atmospheric and common-path aberrations) as well as using or not CDI probes, DM corrections or errors, etc

TODO
    Add astrogrid pattern functionality from MEDIS0

"""

import numpy as np
from scipy import interpolate, ndimage
from inspect import getframeinfo, stack
import matplotlib.pylab as plt
import proper

from medis.params import sp, tp
from medis.CDI import cdi, config_probe
from medis.optics import check_sampling, apodize_pupil, unwrap_phase_zeros as unwrap_phase
from medis.utils import dprint
# from medis.plot_tools import view_spectra, view_timeseries, quick2D, plot_planes


################################################################################
# Deformable Mirror
################################################################################
def deformable_mirror(wf, WFS_map, iter, previous_output=None, apodize=False, plane_name='', debug=False):
    """
    combine different DM actuator commands into single map to send to prop_dm

    prop_dm needs an input map of n_actuators x n_actuators in units of actuator command height. quick_ao will handle
    the conversion to actuator command height, and the CDI probe must be scaled in cdi.probe_amp in params in
    units of m. Each subroutine is also responsible for creating a map of n_actuators x n_actuators spacing. prop_dm
    handles the resampling of this map onto the wavefront, including the influence function. Its some wizardry that
    happens in c, and presumably it is taken care of so you don't have to worry about it.

    In the call to proper.prop_dm, we apply the flag tp.fit_dm, which switches between two 'modes' of proper's DM
    surface fitting. If FALSE, the DM is driven to the heights specified by dm_map, and the influence function will
    act on these heights to define the final surface shape applied to the DM, which may differ substantially from
    the initial heights specified by dm_map. If TRUE, proper will iterate applying the influence function to the
    input heights, and adjust the heights until the difference between the influenced-map and input map meets some
    proper-defined convergence criterea. Setting tp.fit_dm=TRUE will obviously slow down the code, but will (likely)
    more accurately represent a well-calibrated DM response function.

    much of this code copied over from example from Proper manual on pg 94

    :param wf: single wavefront
    :param WFS_map: wavefront sensor map, should be in units of phase delay
    :param previous_output:
    :param iter: the current index of iteration (which timestep this is)
    :param plane_name: name of plane (should be 'woofer' or 'tweeter' for best functionality)
    :return: nothing is returned, but the probe map has been applied to the DM via proper.prop_dm. DM plane post DM
        application can be saved via the sp.save_list functionality
    """
    assert np.logical_xor(WFS_map is None, previous_output is None)

    # AO Actuator Count from DM Type
    if plane_name == 'tweeter' and hasattr(tp,'act_tweeter'):
        nact = tp.act_tweeter
    elif plane_name == 'woofer' and hasattr(tp,'act_woofer'):
        nact = tp.act_woofer
    else:
        nact = tp.ao_act

    # DM Coordinates
    nact_across_pupil = nact - 2  # number of full DM actuators across pupil (oversizing DM extent)
    dm_xc = (nact / 2)   # The location of the optical axis (center of the wavefront) on the DM in
    dm_yc = (nact / 2)   # actuator units. First actuator is centered on (0.0, 0.0). The 0.5 is a
    #  parameter introduced/tuned by Rupert to remove weird errors (address this).
    # KD verified this needs to be here or else suffer weird errors 9/19
    # TODO address/remove the 0.5 in DM x,y coordinates

    ############################
    # Creating DM Surface Map
    ############################
    d_beam = 2 * proper.prop_get_beamradius(wf)  # beam diameter
    act_spacing = d_beam / nact_across_pupil  # actuator spacing [m]

    #######
    # AO
    #######
    if previous_output is not None and WFS_map is None:
        dm_map = update_dm(previous_output)
    else:
        dm_map = quick_ao(wf, nact, WFS_map[wf.iw])

    #########
    # Waffle
    #########
    if tp.satelite_speck['apply'] and plane_name is not 'woofer':
        waffle = make_speckle_kxy(tp.satelite_speck['xloc'], tp.satelite_speck['yloc'],
                                  tp.satelite_speck['amp'], tp.satelite_speck['phase'])
        waffle += make_speckle_kxy(tp.satelite_speck['xloc'], -tp.satelite_speck['yloc'],
                                   tp.satelite_speck['amp'], tp.satelite_speck['phase'])
        dm_map += waffle

    #######
    # CDI
    ######
    if cdi.use_cdi and plane_name == cdi.which_DM:
        theta = cdi.phase_series[iter]
        if not np.isnan(theta):
            # dprint(f"Applying CDI probe, lambda = {wfo.wsamples[iw]*1e9:.2f} nm")
            probe = config_probe(theta, nact, iw=wf.iw, ib=wf.ib)  # iw and ib only used for plotting, and only if sp.verbose=true
            dm_map = dm_map + probe  # Add Probe to DM map

    #########################
    # Applying Piston Error
    #########################
    if tp.piston_error:
        mean_dm_map = np.mean(np.abs(dm_map))
        var = 1e-4  # 1e-11
        dm_map = dm_map + np.random.normal(0, var, (dm_map.shape[0], dm_map.shape[1]))

    #########################
    # proper.prop_dm
    #########################
    dmap = proper.prop_dm(wf, dm_map, dm_xc, dm_yc, act_spacing, FIT=tp.fit_dm)  #

    if debug and wf.iw == 0 and wf.ib == 0:
        dprint(plane_name)
        check_sampling(wf, iter, plane_name+' DM pupil plane', getframeinfo(stack()[0][0]), units='mm')

        # plt.ion()
        fig, ax = plt.subplots(1,1)
        cax = ax.imshow(dm_map*1e9, interpolation='none', origin='lower')
        plt.title(f'{plane_name} dm_map (actuator coordinates)')
        cb = plt.colorbar(cax)
        cb.set_label('nm')
        plt.show(block=True)

        post_ao = unwrap_phase(proper.prop_get_phase(wf)) * wf.lamda / (2 * np.pi)
        # quick2D(pre_ao_dist*1e9, title='unwrapped wavefront before DM', zlabel='nm', show=False)  # , vlim=(-0.5e-7,0.5e-7))
        # quick2D(np.abs(pre_ao_amp)**2, title='Pre-AO Intensity', show=False)#, vlim=(-0.5e-7,0.5e-7))
        # quick2D(dmap, title='the phase map prop_dm is applying', zlabel='distance (m)', show=False)#, vlim=(-0.5e-7,0.5e-7))
        # plt.figure()
        # plt.plot(pre_ao_dist[len(pre_ao_dist)//2], label=f'pre_ao 1D cut, row {len(pre_ao_dist)//2}')
        # plt.plot(2*dmap[len(dmap)//2], label=f'dmap 1D cut (x2), row {len(dmap)//2}')
        # plt.plot((pre_ao_dist + (2*dmap))[len(dmap)//2], label='difference')
        # plt.legend()
        # plt.xlim(sp.grid_size//2*np.array([1-sp.beam_ratio*1.1, 1+sp.beam_ratio*1.1]))
        # quick2D(pre_ao + (2*dmap), title='diff', zlabel='m', show=False, vlim=(-0.5e-7,0.5e-7))
        # quick2D(post_ao, title='unwrapped wavefront after DM', zlabel='m', show=False, vlim=(-0.5e-7,0.5e-7))
        # quick2D(np.abs(proper.prop_get_amplitude(wf))**2, title='wavefront after DM intensity', show=False)
        # quick2D(proper.prop_get_phase(wf), title='wavefront after DM in phase units', zlabel='Phase',
        #          show=True)  # colormap='sunlight',

    if apodize:
        hardmask_pupil(wf)

    return dmap


################################################################################
# Ideal AO
################################################################################
def quick_ao(wf, nact, WFS_map):
    """
    calculate the offset map to send to the DM from the WFS map

    The main idea is to apply the DM only to the region of the wavefront that contains the beam. The phase map from
    the wfs saved the whole wavefront, so that must be cropped. During the wavefront initialization in
    wavefront.initialize_proper, the beam ratio set in sp.beam_ratio is scaled per wavelength (to achieve constant
    sampling sto create white light images), so the cropped value must also be scaled by wavelength. Note, beam ratio
    is scaled differently depending on if sp.focused_sys is True or not. See params-->sp.focused_sys and Proper
    manual pg 36 for more info.

    Then, we interpolate the cropped beam onto a grid of (n_actuators,n_actuators), such that the DM can apply a
    actuator height to each represented actuator, not a over or sub-sampled form. If the number of actuators is low
    compared to the number of samples on the beam, you should anti-alias the WFS map via a lowpass filter before
    interpolating. There is a discrepancy between the sampling of the wavefront at this location (the size you cropped)
    vs the size of the DM. proper.prop_dm handles this, so just plug in the n_actuator sized DM map with specified
    parameters, and assume that prop_dm handles the resampling correctly via the spacing or n_act_across_pupil flag.
    FYI the resampling is done via a c library you installed/compiled when installing proper.

    The WFS map is a map of real values in units of phase delay in radians. However, the AO map that gets passed to
    proper.prop_dm wants input in nm height of each actuator. Therefore, you need to convert the phase delay to
    a DM height. For the ideal AO, you would do this individually for each wavelength. However, for a 'real' AO system
    you do this for the median wavelength. You also need to account for a factor of 2, since the DM is modeled as
    a mirror so it travels the length of the phase delay twice.

    much of this code copied over from example from Proper manual on pg 94

    :param wfo: wavefront object created by optics.Wavefronts() [n_wavelengths, n_objects] of tp.gridsize x tp.gridsize
    :param WFS_map: returned from quick_wfs (as of Aug 2019, its an idealized image)
    :return: ao_map: map of DM actuator command heights in units of m
    """

    nact_across_pupil = nact-2          # number of full DM actuators across pupil (oversizing DM extent)
                                        # Note: oversample by 2 actuators hardcoded here, check if this is appropriate 
    
    ############################
    # Creating AO Surface Map
    ############################
    d_beam = 2 * proper.prop_get_beamradius(wf)  # beam diameter
    act_spacing = d_beam / nact_across_pupil  # actuator spacing [m]

    ###################################
    # Cropping the Beam from WFS map
    ###################################
    # cropping here by beam_ratio rather than d_beam is valid since the beam size was initialized
    #  using the scaled beam_ratios when the wfo was created
    ao_map = WFS_map[
             # sp.grid_size//2 - np.int_(wf.beam_ratio*sp.grid_size//2)+1:
             # sp.grid_size//2 + np.int_(wf.beam_ratio*sp.grid_size//2)+2,
             # sp.grid_size//2 - np.int_(wf.beam_ratio*sp.grid_size//2)+1:
             # sp.grid_size//2 + np.int_(wf.beam_ratio*sp.grid_size//2)+2]
             sp.grid_size//2 - np.int_(wf.beam_ratio*sp.grid_size//2)-1:
             sp.grid_size//2 + np.int_(wf.beam_ratio*sp.grid_size//2)+2,
             sp.grid_size//2 - np.int_(wf.beam_ratio*sp.grid_size//2)-1:
             sp.grid_size//2 + np.int_(wf.beam_ratio*sp.grid_size//2)+2]
    # dprint(f"WFS map coordinates are {sp.grid_size//2 - np.int_(wf.beam_ratio*sp.grid_size//2)-1},"
    #        f"{sp.grid_size//2 + np.int_(wf.beam_ratio*sp.grid_size//2)+1}")

    ########################################################
    # Interpolating the WFS map onto the actuator spacing
    # (tp.nact,tp.nact)
    ########################################################
    # Lowpass Filter- prevents aliasing; uses Gaussian filter
    nyquist_dm = tp.ao_act/2 * act_spacing  # [m]
    sigma = [nyquist_dm/2.355, nyquist_dm/2.355]  # assume we want sigma to be twice the HWHM
    ao_map = ndimage.gaussian_filter(ao_map, sigma=sigma, mode='nearest')

    f = interpolate.interp2d(range(ao_map.shape[0]), range(ao_map.shape[0]), ao_map, kind='cubic')
    ao_map = f(np.linspace(0,ao_map.shape[0],nact), np.linspace(0,ao_map.shape[0], nact))
    # map_spacing = proper.prop_get_sampling(wf)
    # ao_map = proper.prop_magnify(ao_map, map_spacing / act_spacing, nact, QUICK=True)

    ################################################
    # Converting phase delay to DM actuator height
    ################################################
    # Apply the inverse of the WFS image to the DM, so use -dm_map (dm_map is in phase units, divide by k=2pi/lambda)
    surf_height = proper.prop_get_wavelength(wf) / (4 * np.pi)  # [m/rad]
    ao_map = -ao_map * surf_height  # Converts DM map to units of [m] of actuator heights

    return ao_map


def retro_wfs(star_fields, wfo, plane_name='wfs'):
    """
    Retrospective wfs (measure an old field)

    :param star_fields:
    :param wfo:
    :param plane_name:
    :return:
    """
    WFS_map = np.zeros((len(star_fields), sp.grid_size, sp.grid_size))
    from skimage.restoration import unwrap_phase
    for iw in range(len(star_fields)):
        quick2D(np.angle(star_fields), title='before mask', colormap='sunlight')
        phasemap = np.angle(star_fields[iw])
        masked_phase = np.ma.masked_equal(phasemap, 0)
        quick2D(masked_phase, title='before unwrap', colormap='sunlight')
        WFS_map[iw] = unwrap_phase(masked_phase, wrap_around=[False, False])
        WFS_map[iw][phasemap == 0] = 0
        quick2D(WFS_map[iw], title='after')
    if 'retro_closed_wfs' in sp.save_list:
        wfo.save_plane(location='WFS_map')

    return WFS_map


def open_loop_wfs(wfo, plane_name='wfs'):
    """
    saves the unwrapped phase [arctan2(imag/real)] of the wfo.wf_collection at each wavelength

    It is an idealized image (exact copy) of the wavefront phase per wavelength. Only the map for the first object
    (the star) is saved. We have initialized

    Here we hardmask on the WFS map to be a circle around the beam in the pupil plane. This hard masking prevents the
     DM from acting on non-beam signal, since the DM modelled by proper is a nxn square array, but the beam is nominally
     circular for circular apertures.

    #TODO the way this is saved for naming the WFS_map is going to break if you want to do closed loop WFS on a
    #TODO woofer-tweeter system

    :param wfo: wavefront object
    :param plane_name: name of the plane to enable or disable saving the WFS map
    :return: array containing only the unwrapped phase delay of the wavefront; shape=[n_wavelengths], units=radians
    """
    from skimage.restoration import unwrap_phase

    star_wf = wfo.wf_collection[:, 0]
    WFS_map = np.zeros((len(star_wf), sp.grid_size, sp.grid_size))

    for iw in range(len(star_wf)):
        hardmask_pupil(star_wf[iw])
        phasemap = proper.prop_get_phase(star_wf[iw])
        WFS_map[iw] = unwrap_phase(phasemap, wrap_around=[False, False])
        WFS_map[iw][phasemap==0] = 0

    if 'open_loop_wfs' in sp.save_list or sp.closed_loop:
        wfo.save_plane(location='WFS_map')

    return WFS_map


def hardmask_pupil(wf):
    """
    hard-edged circular mask of the pupil plane.

    Masks out the WFS map outside of the beam since the DM modeled by proper can only be a square nxn grid of actuators,
    and thus the influence function surrounding each DM actuator could  be affecting on-beam pixels, even if the
    actuator is acting on off-beam signal. In other words, even if a cornerDM actuator doesn't actuate on the beam,
    if there was non-zero signal in the WFS map, it will try to act on it, and it could 'influence' nearby DM actuators
    that are acting on the beam.

    This hard-edged mask is different from prop_circular_aperture in that it does not anti-alias the edges of the mask
    based on the 'fill factor' of the edge pixels. Instead, it has a boolean mask to zero everything > a fixed radius,
    in this case determined by the grid size and beam ratio of each wavefront passed into it.

    :param wf: a single wavefront
    :return: nothing is returned but the wf passed into it has been masked

    """
    phase_map = proper.prop_get_phase(wf)
    amp_map = proper.prop_get_amplitude(wf)

    # Sizing the Mask
    h, w = wf.wfarr.shape[:2]
    center = (int(w / 2), int(h / 2))
    radius = np.floor(sp.grid_size * wf.beam_ratio / 2)  # Should scale with wavelength if sp.focused_system=False,
                                                        # np.ceil used to oversize map so don't clip the beam
    # Making the Circular Boolean Mask
    Y, X = np.mgrid[:h, :w]
    dist_from_center = np.sqrt((X - center[0]) ** 2 + (Y - center[1]) ** 2)
    inds = dist_from_center <= radius

    # Applying the Mask to the Complex Array
    mask = np.zeros_like(phase_map)
    mask[inds] = 1
    masked = phase_map * mask
    wf.wfarr = proper.prop_shift_center(amp_map * np.cos(masked) + 1j * amp_map * np.sin(masked))

    if sp.debug and wf.ib == 0:
        dprint(f"Radius of hard-edge pupil mask is {radius} pixels")
        from medis.plot_tools import quick2D
        quick2D(masked, title=f"Masked phase map in hardmask_pupil, lambda={wf.lamda*1e9} nm", zlabel='phase?')


def make_speckle_kxy(kx, ky, amp, dm_phase):
    """given an kx and ky wavevector,
    generates a NxN flatmap that has
    a speckle at that position"""
    N = tp.ao_act
    dmx, dmy   = np.meshgrid(
                    np.linspace(-0.5, 0.5, N),
                    np.linspace(-0.5, 0.5, N))

    xm=dmx*kx*2.0*np.pi
    ym=dmy*ky*2.0*np.pi
    # print 'DM phase', dm_phase
    ret = amp*np.cos(xm + ym +  dm_phase)
    return ret


# def ao(wf, WFS_map, theta):
#     if sp.closed_loop:
#         deformable_mirror(wf, WFS_map, theta)
#     else:
#         WFS_map = open_loop_wfs(wf)  # overwrite WFS_map
#         # dprint(f"WFS_ma.shape = {WFS_map.shape}")
#         deformable_mirror(wf, WFS_map, theta)
#
################################################################################
# Full AO
################################################################################
# not implemented. Full AO implies a time delay, and maybe non-ideal WFS



"""
check the params entered to params for given telescope settings for correct sampling
KD 8/13/19

According to the Proper manual:
    sp.beam_ratio = tp.entrance_d/ init_grid_width
    where init_grid_width is the initial grid width (diameter) in meters
Thus
    init_grid_width = tp.enterence_d / sp.beam_ratio

The sampling of the beam is defined by
    sampling = grid_width / sp.grid_size
    :keyword
    sampling = tp.enterence_d * sp.grid_size / sp.beam_ratio
    where grid_width here is the initial grid width

"""
import numpy as np
from medis.params import tp, sp

def check_sampling():
    #######################################
    # Definitions from inside mini-medis
    #######################################
    # Basics-not imported
    # tp.entrance_d = 0.1  # [m]
    # sp.beam_ratio = .3  # [unitless]
    # sp.grid_size = 512  # [pix]

    # From Params Buried in mini-medis
    legs_frac = 0.1  # [m]

    #######################
    # Initial Sampling
    ######################
    init_grid_width = tp.entrance_d / sp.beam_ratio  # [m]
    init_sampling = init_grid_width / sp.grid_size  # [m/pix]
    init_npix_on_beam = tp.entrance_d / init_sampling  # [pix]
    print(f"Initial Grid Width = {init_grid_width:.2f} m")
    print(f"Initial number of pixels across beam = {sp.beam_ratio*sp.grid_size}")

    #################################
    # Checking Sampling of Spiders
    ################################
    spider_width = tp.entrance_d * legs_frac  # [m]
    npix_on_spiders = spider_width / init_sampling  # [pix]
    print(f"Number of pixels across the spiders = {npix_on_spiders}")
    print("********************************")

###################
# Ideal Sampling
###################
def ideal_sampling():
    """
    see proper manual pg 36 for definitions

    init_grid_width is dependant on chosen/user-defined beam ratio. In order to determine what a "good" one would be,
    first choose a grid_width that is significantly larger than the entrance diameter, such that the beam is only
    resolved by a few pixels. This ensures there is sufficient zero-padding of the pupil to avoid FFT effects, and you
    can check the sampling of the spiders with the code above. There is a trade-off between sampling in the pupil and
    the image planes, so it is best to only choose a beam ratio that barely resolves the smallest features of the aperture
    (spiders), in order to get the best resolution of the image plane.
    """
    # Hard-Coded Parameters
    Fnum = 12.6  # focal ratio of the beam as it approaches the focus
    lmbda = 800e-9  # [m]  wavelength of interest
    grid_width = tp.entrance_d/sp.beam_ratio  # [m]

    # Sampling
    samp_good = Fnum*lmbda*tp.entrance_d / grid_width
    samp_nyq = Fnum*lmbda/2
    print(f"In the pupil plane at lambda={lmbda*1e6:.3f} um:\n"
          f" Ideal Sampling is {samp_good:.6f} m/pix \n"
          f"Nyquist Sampling is {samp_nyq:.6f} m/pix ")

    # selecting beam ratio
    br_good = tp.entrance_d/(samp_good*sp.grid_size)
    br_nyq = tp.entrance_d/samp_nyq/sp.grid_size
    print(f"Good starter beam ratio is {br_good} \n"
          f"Nyquist beam ratio is {br_nyq}")

    ###########################################
    # Beam Size and Grid Width at other Plane
    ###########################################
    final_sampling = 0.212727  # m/pix
    final_grid_width = final_sampling * sp.grid_size  # m
    final_beam_size = final_grid_width * sp.beam_ratio  # m
    final_npix_on_beam = final_beam_size / final_sampling  # pixels


if __name__ == '__main__':
    check_sampling()



"""
run_SubaruFrontend
KD
Dec 9 2019

This is the starting point to run the Subaru_frontend prescription. From here, you can turn on/off toggles, change AO
settings, view different planes, and make other changes to running the prescription without changing the base
prescription or the default params themselves.

This file is meant to 'cleanup' a lot of the functionality between having defaults saved in the params file and
the prescription. Most changes necessary to run multiple instances of the base prescription will be made here. This is
the same layot/format as the Example files that Rupert was using in v1.0. These are now meant to be paired with a more
specific prescription than the original optics_propagate, which had more toggles and less sophisiticated optical train.

"""
import numpy as np


from medis.params import iop, sp, ap, tp
from medis.utils import dprint
import medis.optics as opx
from medis.plot_tools import view_spectra, view_timeseries, quick2D, plot_planes
import medis.medis_main as mm

#################################################################################################
#################################################################################################
#################################################################################################
# Renaming Directories
testname = 'SubaruAO188-test1'
iop.update_datadir('/home/captainkay/mazinlab/MKIDSim/CDIsim_data/')
iop.update_testname(testname)
iop.makedir()

# Telescope
tp.prescription = 'Subaru_frontend'
tp.entrance_d = 7.9716
tp.flen_primary = tp.entrance_d * 13.612

# Simulation & Timing
sp.numframes = 1
sp.closed_loop = True

# Grid Parameters
sp.focused_sys = True
sp.beam_ratio = 0.08  # parameter dealing with the sampling of the beam in the pupil/focal plane
sp.grid_size = 512  # creates a nxn array of samples of the wavefront
sp.maskd_size = 256  # will truncate grid_size to this range (avoids FFT artifacts) # set to grid_size if undesired

# Companion
ap.companion = False
ap.contrast = [1e-1]
ap.companion_xy = [[5, -5]]  # units of this are lambda/tp.entrance_d

# Toggles for Aberrations and Control
tp.obscure = False
tp.use_atmos = False
tp.use_aber = False
tp.use_ao = True
tp.act_woofer = 16  # A 14x14 grid gives the closest nact to 188 (169 act). But, we oversize the DM by one actuator in
                    # each direction in the DM code, so a 16x16 grid gives a roughly 14x14 grid over the beam

# Plotting
sp.show_wframe = True  # Plot white light image frame
sp.show_spectra = False  # Plot spectral cube at single timestep
sp.spectra_cols = 3  # number of subplots per row in view_spectra
sp.show_tseries = False  # Plot full timeseries of white light frames
sp.tseries_cols = 5  # number of subplots per row in view_timeseries
sp.show_planes = False
sp.debug = False
sp.debug_planes = False

# Saving
sp.save_to_disk = False  # save obs_sequence (timestep, wavelength, x, y)
sp.save_list = ['woofer', 'detector']  # list of locations in optics train to save


if __name__ == '__main__':
    # =======================================================================
    # Run it!!!!!!!!!!!!!!!!!
    # =======================================================================
    sim = mm.RunMedis(name=testname, product='fields')

    observation = sim()
    cpx_sequence = observation['fields']
    sampling = observation['sampling']

    # =======================================================================
    # Focal Plane Processing
    # =======================================================================
    # obs_sequence = np.array(obs_sequence)  # obs sequence is returned by gen_timeseries (called above)
    # (n_timesteps ,n_planes, n_waves_init, n_astro_bodies, nx ,ny)
    cpx_sequence = opx.interp_wavelength(cpx_sequence, 2)  # interpolate over wavelength
    focal_plane = opx.extract_plane(cpx_sequence, 'detector')  # eliminates object axis
    # convert to intensity THEN sum over object, keeping the dimension of tstep even if it's one
    focal_plane = np.sum(opx.cpx_to_intensity(focal_plane), axis=2)
    fp_sampling = np.copy(sampling[cpx_sequence.shape[1]-1,:])
            # numpy arrays have some weird effects that make copying the array necessary

    # =======================================================================
    # Plotting
    # =======================================================================
    # White Light, Last Timestep
    if sp.show_wframe:
        # vlim = (np.min(spectralcube) * 10, np.max(spectralcube))  # setting z-axis limits
        img = np.sum(focal_plane[sp.numframes-1], axis=0)  # sum over wavelength
        quick2D(opx.extract_center(img), #focal_plane[sp.numframes-1]),
                title=f"White light image at timestep {sp.numframes} \n"  # img
                           f"AO={tp.use_ao}",
                           # f"Grid Size = {sp.grid_size}, Beam Ratio = {sp.beam_ratio} ",
                           # f"sampling = {sampling*1e6:.4f} (um/gridpt)",
                logZ=True,
                dx=fp_sampling[0],
                vlim=(1e-7, 1e-3))

    # Plotting Spectra at last tstep
    if sp.show_spectra:
        tstep = sp.numframes-1
        view_spectra(focal_plane[sp.numframes-1],
                      title=f"Intensity per Spectral Bin at Timestep {tstep} \n"
                            f" AO={tp.use_ao}",
                            # f"Beam Ratio = {sp.beam_ratio:.4f}",#  sampling = {sampling*1e6:.4f} [um/gridpt]",
                      logZ=True,
                      subplt_cols=sp.spectra_cols,
                      vlim=(1e-4, 1e-1),
                      dx=fp_sampling)

    # Plotting Timeseries in White Light
    if sp.show_tseries:
        img_tseries = np.sum(focal_plane, axis=1)  # sum over wavelength
        view_timeseries(img_tseries, title=f"White Light Timeseries\n"
                                            f"AO={tp.use_ao}",
                        subplt_cols=sp.tseries_cols,
                        logZ=True,
                        vlim=(1e-6, 1e-3))
                        # dx=fp_sampling[0])

    # Plotting Selected Plane
    if sp.show_planes:
        vlim = [(None, None), (None, None), (None, None), (None, None), (None, None)]
        # vlim = [(None,None), (None,None), (None,None), (1e-2,1e-1)]  # (7e-4, 6e-4)
        logZ = [True, False, False, True]
        if sp.save_list:
            plot_planes(cpx_sequence,
                        title=f"White Light through Optical System",
                        subplt_cols=2,
                        vlim=vlim,
                        logZ=logZ,
                        dx=sampling)

    test = 1



"""
run_SCExAO
KD
Feb 2020

This is the starting point to run the Subaru_SCExAO prescription. From here, you can turn on/off toggles, change AO
settings, view different planes, and make other changes to running the prescription without changing the base
prescription or the default params themselves.

"""
import numpy as np

from medis.params import sp, tp, iop, ap
from medis.CDI import cdi, cdi_postprocess
from medis.utils import dprint
import medis.optics as opx
from medis.plot_tools import view_spectra, view_timeseries, quick2D, plot_planes
import medis.medis_main as mm

#################################################################################################
#################################################################################################
#################################################################################################
testname = 'SCExAO-test1'
iop.update_datadir(f"/home/captainkay/mazinlab/MKIDSim/CDIsim_data/")
iop.update_testname(testname)
iop.makedir()

# Telescope
tp.prescription = 'Subaru_SCExAO'
tp.entrance_d = 7.9716
tp.flen_primary = tp.entrance_d * 13.612

# Simulation & Timing
sp.numframes = 1
sp.closed_loop = False

# Grid Parameters
sp.focused_sys = True
sp.beam_ratio = 0.08  # parameter dealing with the sampling of the beam in the pupil/focal plane
sp.grid_size = 512  # creates a nxn array of samples of the wavefront
sp.maskd_size = 256  # will truncate grid_size to this range (avoids FFT artifacts) # set to grid_size if undesired

# Companion
ap.companion = True
ap.contrast = [5e-2, 9e-3]
ap.companion_xy = [[5, -5], [8, 9.5]]  # units of this are lambda/tp.entrance_d
ap.n_wvl_init = 3  # initial number of wavelength bins in spectral cube (later sampled by MKID detector)
ap.n_wvl_final = None  # final number of wavelength bins in spectral cube after interpolation (None sets equal to n_wvl_init)
ap.interp_wvl = False  # Set to interpolate wavelengths from ap.n_wvl_init to ap.n_wvl_final
ap.wvl_range = np.array([800, 1400]) / 1e9  # wavelength range in [m] (formerly ap.band)
# eg. DARKNESS band is [800, 1500], J band =  [1100,1400])

# CDI
cdi.use_cdi = False
cdi.probe_w = 10  # [actuator coordinates] width of the probe
cdi.probe_h = 30  # [actuator coordinates] height of the probe
cdi.probe_center = (10,10)  # [actuator coordinates] center position of the probe
cdi.probe_amp = 2e-8  # [m] probe amplitude, scale should be in units of actuator height limits
cdi.which_DM = 'tweeter'
cdi.phs_intervals = np.pi/2
cdi.phase_integration_time = 0.01


# Toggles for Aberrations and Control
tp.obscure = False
tp.use_atmos = True
tp.use_aber = False
tp.add_zern = False  # Just a note: zernike aberrations generate randomly each time the telescope is run, so introduces
                     # potentially inconsistent results
tp.use_ao = True
sp.skip_functions = []  # skip_functions is based on function name, alternate way of on/off than the toggling
                    # 'coronagraph' 'deformable_mirror' 'add_aber'
# Plotting
sp.show_wframe = True  # plot white light image frame
sp.show_spectra = True  # Plot spectral cube at single timestep
sp.spectra_cols = 3  # number of subplots per row in view_spectra
sp.show_tseries = False  # Plot full timeseries of white light frames
sp.tseries_cols = 3  # number of subplots per row in view_timeseries
sp.show_planes = True
sp.maskd_size = 256
sp.verbose = False

# Saving
sp.save_to_disk = False  # save obs_sequence (timestep, wavelength, x, y)
sp.save_list = [ 'entrance_pupil','woofer', 'tweeter',   'detector']  # list of locations in optics train to save 'entrance_pupil',
                # 'entrance_pupil','post-DM-focus', 'coronagraph',

if __name__ == '__main__':
    # =======================================================================
    # Run it!!!!!!!!!!!!!!!!!
    # =======================================================================
    sim = mm.RunMedis(name=testname, product='fields')
    observation = sim()
    cpx_sequence = observation['fields']
    sampling = observation['sampling']

    # =======================================================================
    # Focal Plane Processing
    # =======================================================================
    # cpx_sequence = (n_timesteps ,n_planes, n_waves_init, n_astro_bodies, nx ,ny)
    cpx_sequence = opx.interp_wavelength(cpx_sequence, 2)  # interpolate over wavelength
    focal_plane = opx.extract_plane(cpx_sequence, 'detector')  # eliminates astro_body axis
    # convert to intensity THEN sum over object, keeping the dimension of tstep even if it's one
    focal_plane = np.sum(opx.cpx_to_intensity(focal_plane), axis=2)
    fp_sampling = np.copy(sampling[cpx_sequence.shape[1]-1,:])  # numpy arrays have some weird effects that make copying the array necessary

    # ======================================================================
    # CDI
    # ======================================================================
    # if cdi.use_cdi:
        # cdi_postprocess(cpx_sequence, fp_sampling, plot=True)
        #     cdi.save_tseries(img_tseries)
        #     cdi.save_cout_to_disk()

    # =======================================================================
    # Plotting
    # =======================================================================
    # White Light, Last Timestep
    if sp.show_wframe:
        # vlim = (np.min(spectralcube) * 10, np.max(spectralcube))  # setting z-axis limits
        img = np.sum(focal_plane[sp.numframes-1], axis=0)  # sum over wavelength
        quick2D(opx.extract_center(img), #focal_plane[sp.numframes-1]),
                title=f"White light image at timestep {sp.numframes} \n"  # img
                           f"AO={tp.use_ao}, CDI={cdi.use_cdi} ",
                           # f"Grid Size = {sp.grid_size}, Beam Ratio = {sp.beam_ratio} ",
                           # f"sampling = {sampling*1e6:.4f} (um/gridpt)",
                logZ=True,
                dx=fp_sampling[0],
                vlim=(None,None))  # (1e-3, 1e-1)

    # Plotting Spectra at last tstep
    if sp.show_spectra:
        tstep = sp.numframes-1
        view_spectra(focal_plane[sp.numframes-1],
                      title=f"Intensity per Spectral Bin at Timestep {tstep} \n"
                            f" AO={tp.use_ao}, CDI={cdi.use_cdi}",
                            # f"Beam Ratio = {sp.beam_ratio:.4f}",#  sampling = {sampling*1e6:.4f} [um/gridpt]",
                      logZ=True,
                      subplt_cols=sp.spectra_cols,
                      vlim=(1e-7, 1e-3),
                      dx=fp_sampling)

    # Plotting Timeseries in White Light
    if sp.show_tseries:
        img_tseries = np.sum(focal_plane, axis=1)  # sum over wavelength
        view_timeseries(img_tseries, title=f"White Light Timeseries\n"
                                            f"AO={tp.use_ao}. CDI={cdi.use_cdi}",
                        subplt_cols=sp.tseries_cols,
                        logZ=True,
                        vlim=(1e-10, 1e-4),
                        dx=fp_sampling[0])

    # Plotting Selected Plane
    if sp.show_planes:
        vlim = [(None, None), (None, None), (None, None), (1e-7,1e-3), (1e-7,1e-3), (1e-7,1e-3)]
        # vlim = [(None,None), (None,None), (None,None), (None,None)]  # (1e-2,1e-1) (7e-4, 6e-4)
        logZ = [True, False, False, True, True, True]
        if sp.save_list:
            plot_planes(cpx_sequence,
                        title=f"White Light through Optical System",
                        subplt_cols=2,
                        vlim=vlim,
                        logZ=logZ,
                        dx=sampling)

    test = 1



import numpy as np
from inspect import getframeinfo, stack
import pickle
import tables as pt
import astropy.io.fits as afits

from medis.params import sp, ap, tp, iop


def dprint(*message, path_display=-3):
    """
    prints location of code where message is printed from

    >>> dprint('foo', 5000, (), np.arange(9).reshape(3,3))
    MEDIS++/medis/optics.py:173 - lol, 5000, (), [[0 1 2]
    [3 4 5]
    [6 7 8]]

    path_to_display : integer number of folders back from the module location to display in printed statement
    """
    caller = getframeinfo(stack()[1][0])
    message_str = ''
    for mess in message:
        message_str += f'{mess}, '
    message_str = message_str[:-2]

    reduced_filename = '/'.join(caller.filename.split('/')[path_display:])
    print("%s:%d - %s" % (reduced_filename, caller.lineno, message_str))


def phase_cal(wavelengths):
    """Wavelength in nm"""
    phase = tp.wavecal_coeffs[0] * wavelengths + tp.wavecal_coeffs[1]
    return phase


####################################################################################################
# Functions Relating to Reading, Loading, and Saving Data #
####################################################################################################
def save_to_disk_sequence(obs_sequence, obs_seq_file='obs_seq.pkl'):
    """saves obs sequence as a .pkl file

    :param obs_sequence- Observation sequence, 6D data structure
    :param obs_seq_file- filename for saving, including directory tree
    """
    #dprint((obs_seq_file, obs_seq_file[-3:], obs_seq_file[-3:] == '.h5'))
    if obs_seq_file[-3:] == 'pkl':
        with open(obs_seq_file, 'wb') as handle:
            pickle.dump(obs_sequence, handle, protocol=pickle.HIGHEST_PROTOCOL)
    elif obs_seq_file[-3:] == 'hdf' or obs_seq_file[-3:] == '.h5':
        f = pt.open_file(obs_seq_file, 'w')
        ds = f.create_array(f.root, 'data', obs_sequence)
        f.close()
    else:
        dprint('Extension not recognised')


def check_exists_obs_sequence(plot=False):
    """
    This code checks to see if there is already
    an observation sequence saved with the output of the run in the
    location specified by the iop.

    :return: boolean flag if it can find a file or not
    """
    import os
    if os.path.isfile(iop.obs_seq):
        dprint(f"File already exists at {iop.obs_seq}")
        return True
    else:
        return False


def open_obs_sequence(obs_seq_file='hyper.pkl'):
    """opens existing obs sequence .pkl file and returns it"""
    with open(obs_seq_file, 'rb') as handle:
        obs_sequence =pickle.load(handle)
    return obs_sequence


def open_obs_sequence_hdf5(obs_seq_file='hyper.h5'):
    """opens existing obs sequence .h5 file and returns it"""
    # hdf5_path = "my_data.hdf5"
    read_hdf5_file = pt.open_file(obs_seq_file, mode='r')
    # Here we slice [:] all the data back into memory, then operate on it
    obs_sequence = read_hdf5_file.root.data[:]
    # hdf5_clusters = read_hdf5_file.root.clusters[:]
    read_hdf5_file.close()
    return obs_sequence


####################################################################################################
# Functions Relating to Reading, Loading, and Saving Images #
####################################################################################################
def saveFITS(image, name='test.fit'):
    header = afits.Header()
    header["PIXSIZE"] = (0.16, " spacing in meters")

    hdu = afits.PrimaryHDU(image, header=header)
    hdu.writeto(name)


def readFITS(filename):
    """
    reads a fits file and returns data fields only

    :param filename: must specify full filepath
    """
    hdulist = afits.open(filename)
    header = hdulist[0].header
    scidata = hdulist[0].data

    return scidata


def clipped_zoom(img, zoom_factor, **kwargs):
    from scipy.ndimage import zoom
    h, w = img.shape[:2]

    # For multichannel images we don't want to apply the zoom factor to the RGB
    # dimension, so instead we create a tuple of zoom factors, one per array
    # dimension, with 1's for any trailing dimensions after the width and height.
    zoom_tuple = (zoom_factor,) * 2 + (1,) * (img.ndim - 2)

    # Zooming out
    if zoom_factor < 1:

        # Bounding box of the zoomed-out image within the output array
        zh = int(np.round(h * zoom_factor))
        zw = int(np.round(w * zoom_factor))
        top = (h - zh) // 2
        left = (w - zw) // 2

        # Zero-padding
        out = np.zeros_like(img)
        out[top:top+zh, left:left+zw] = zoom(img, zoom_tuple, **kwargs)

    # Zooming in
    elif zoom_factor > 1:

        # Bounding box of the zoomed-in region within the input array
        zh = int(np.round(h / zoom_factor))
        zw = int(np.round(w / zoom_factor))
        top = (h - zh) // 2
        left = (w - zw) // 2
        from medis.Utils.plot_tools import quicklook_im
        out = zoom(img[top:top+zh, left:left+zw], zoom_tuple, **kwargs)
        # quicklook_im(out, logZ=True)
        # `out` might still be slightly larger than `img` due to rounding, so
        # trim off any extra pixels at the edges
        trim_top = ((out.shape[0] - h) // 2)
        trim_left = ((out.shape[1] - w) // 2)
        # print top, zh, left, zw
        # print out.shape[0], trim_top, h, trim_left, w
        if trim_top < 0 or trim_left < 0:
            temp = np.zeros_like(img)
            temp[:out.shape[0],:out.shape[1]] = out
            out = temp
        else:
            out = out[trim_top:trim_top+h, trim_left:trim_left+w]
        # quicklook_im(out, logZ=False)
    # If zoom_factor == 1, just return the input array
    else:
        out = img

    # import matplotlib.pyplot as plt
    # plt.hist(out.flatten(), bins =100, alpha =0.5)
    # plt.hist(img.flatten(), bins =100, alpha=0.5)
    # plt.show()

    # print(np.sum(img), np.sum(out))
    # out = out*np.sum(img)/np.sum(out)
    # out = out*4
    return out



"""Quick plotting tools go here

Note: older code from Rupert used pylab, which is now discouraged. KD changed to pyplot on 7-10-19 but has
not tested all older aspects of the code to ensure proper switch from pylab (though should be the same)
"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm, SymLogNorm
import matplotlib.ticker as ticker
import matplotlib.gridspec as gridspec
from mpl_toolkits.axes_grid1 import ImageGrid
import warnings

from medis.params import tp, sp, ap
from medis.CDI import cdi
from medis.utils import dprint
import medis.optics as opx
from medis.twilight_colormaps import sunlight

# MEDIUM_SIZE = 17
# plt.rc('font', size=MEDIUM_SIZE)  # controls default text sizes

from matplotlib import rcParams
# rcParams['text.usetex'] = False
rcParams['font.family'] = 'DejaVu Sans'
# rcParams['mathtext.fontset'] = 'custom'
# rcParams['mathtext.fontset'] = 'stix'
# rcParams['mathtext.rm'] = 'Bitstream Vera Sans'
# rcParams['mathtext.bf'] = 'Bitstream Vera Sans:bold'


def quick2D(image, dx=None, title=None, logZ=False, vlim=(None,None),
            colormap='YlGnBu_r', zlabel='Intensity', show=True):
    """
    Looks at a 2D array, has bunch of handles for plot.imshow

    :param image: 2D array to plot (data)
    :param dx: sampling of the image in m. Hardcoded to convert to um on axis
    :param title: string--must be set or will error!
    :param logZ: flag to set logscale plotting on z-axis
    :param vlim: tuple of limits on the colorbar axis, otherwise default matplotlib (pass in logscale limits if logZ=True)
    :param colormap: specify colormap as string
    :return:
    """
    # Create figure & adjust subplot number, layout, size, whitespace
    fig = plt.figure()
    ax = fig.add_subplot(111)

    # Title
    while title is None:
        warnings.warn("Plots without titles: Don't Do It!")
        title = input("Please Enter Title: ")

    # X,Y lables
    if dx is not None:
        scale = np.round(
            np.linspace(-dx * sp.maskd_size / 2, dx * sp.maskd_size / 2, (sp.maskd_size + 1) / 50) * 1e6).astype(int)
        tic_spacing = np.linspace(0, sp.maskd_size, sp.maskd_size/50)
        tic_spacing[0] = tic_spacing[0] + 1  # hack for edge effects
        tic_spacing[-1] = tic_spacing[-1] -1  # hack for edge effects
        plt.xticks(tic_spacing, scale)
        plt.yticks(tic_spacing, scale)
        plt.xlabel('[um]')

    # Setting z-axis by mean
    # if vlim == (None, None):
    if vlim == (-1, -1):  # Hack this to have it not be default. Only use it if you want it purposely
        nstd = 2
        std = np.std(image)
        mean = np.mean(image)
        vlim = mean - nstd * std, mean + nstd * std

    # Setting Logscale
    if colormap == 'sunlight':
        colormap = sunlight
    if not logZ:
        norm = None
    elif vlim[0] is None or vlim[0] > 0:
        norm = LogNorm()
    else:
        norm = SymLogNorm()
    # norm = None if not logZ else LogNorm() if vlim[0] > 0 else SymLogNorm(1e-7))
    cax = ax.imshow(image, interpolation='none', origin='lower', vmin=vlim[0], vmax=vlim[1],
                    norm=norm, cmap=colormap)

    # Plotting
    plt.title(title, fontweight='bold', fontsize=16)
    cb = plt.colorbar(cax)
    cb.set_label(zlabel)
    if show:
        plt.show(block=False)


def grid(fields, title='body spectra', logZ=False, show=True, nstd=1, vlim=(None, None), cmap='inferno'):
    """
    General purpose plotter for multi-dimensional input tensors from 2D up to 6D. The tensor will be converted to 4D
    and plot as a grid of 2D images

    :param fields:
    :param title:
    :param logZ:
    :param show:
    :return:
    """
    if cmap == 'sunlight':
        cmap = sunlight
    fields = np.array(fields)  # just in case its a list
    assert fields.ndim > 2
    if np.iscomplexobj(fields.flat[0]):
        fields = np.abs(fields)**2  # convert to intensity if complex
    while len(fields.shape) > 4:
        try:
            boring_ind = fields.shape.index(1)
            fields = np.mean(fields, axis=boring_ind)
        except ValueError:
            fields = fields[0]  # if none are zero slice out first dimension
    while len(fields.shape) < 4:
        fields = fields[:,np.newaxis]

    slices = np.int_(np.ceil(np.array(fields.shape)[:2]/5))
    fields = fields[::slices[0],::slices[1]]
    print(f'fields being sliced by {slices} making new fields size {fields.shape}')
    nwave, nobj, x, y = fields.shape

    try:
        std = np.std(fields)
        mean = np.mean(fields)
    except ValueError:
        std = np.std(fields[0])
        mean = np.mean(fields[0])

    if vlim == (None, None):
        vmin, vmax = mean - nstd*std, mean + nstd*std
    else:
        vmin, vmax = vlim

    fig = plt.figure(figsize=(16, 9))
    fig.suptitle(title)
    norm = None if not logZ else (LogNorm() if vmin > 0 else SymLogNorm(1e-7))

    imgrid = ImageGrid(fig, 111,  # as in plt.subplot(111)
                     nrows_ncols=(nobj, nwave),
                     axes_pad=0.15,
                     share_all=True,
                     cbar_location="right",
                     cbar_mode="single",
                     cbar_size="7%",
                     cbar_pad=0.15,
                     )
    for i, ax in enumerate(imgrid):
        x, y = i % nwave, i // nwave
        im = ax.imshow(fields[x, y], norm=norm, vmin=vmin, vmax=vmax, cmap=cmap)

    ax.cax.colorbar(im)
    ax.cax.toggle_label(True)

    plt.tight_layout()
    plt.subplots_adjust(right=0.9)

    if show:
        plt.show(block=True)


def view_spectra(datacube, title=None, show=True, logZ=False, use_axis=True, vlim=(None,None), subplt_cols=3,
                  dx=None):
    """
    view plot of intensity in each wavelength bin at a single (last) timestep

    :param datacube: 3D spectral cube (n_wavelengths, nx, ny) at single timestep
    :param title: string, must be set or will error!
    :param show: flag possibly useful for plotting loops of things?
    :param logZ: turn logscale plotting for Z axis on or off
    :param use_axis: turn on/off using axis ticks, colorbar, etc
    :param vlim: tuple of colorbar axis limits (min,max)
    :param subplt_cols: number of subplots per row
    :param dx: sampling of the image in m. Hardcoded to convert to um
    :return:
    """
    plt.close('all')

    # Create figure & adjust subplot number, layout, size, whitespace
    fig = plt.figure()
    n_colors = len(datacube)
    n_rows = int(np.ceil(n_colors / float(subplt_cols))+1)
    plt.axis('off')
    gs = gridspec.GridSpec(n_rows, subplt_cols, wspace=0.08, top=0.9)

    # Title
    if title is None:
        warnings.warn("Plots without titles: Don't Do It!")
        title = input("Please Enter Title: ")
        pass
    fig.suptitle(title, fontweight='bold', fontsize=16)

    # Wavelength Strings for Subplot Titles
    w_string = np.array(np.linspace(ap.wvl_range[0] * 1e9, ap.wvl_range[1] * 1e9, ap.n_wvl_final, dtype=int), dtype=str)

    for w in range(n_colors):
        ax = fig.add_subplot(gs[w])
        ax.set_title(r'$\lambda$ = ' + f"{w_string[w]} nm")

        slice = opx.extract_center(datacube[w])

        # X,Y lables
        if dx is not None:
            dx[w] = dx[w] * 1e6  # [convert to um]
            # dprint(f"sampling = {sampl[w]}")
            tic_spacing = np.linspace(0, slice.shape[0], 5)  # 5 (number of ticks) is set by hand, arbitrarily chosen
            tic_lables = np.round(
                np.linspace(-dx[w] * sp.maskd_size / 2, dx[w] * sp.maskd_size / 2, 5)).astype(int)  # nsteps must be same as tic_spacing
            tic_spacing[0] = tic_spacing[0] + 1  # hack for edge effects
            tic_spacing[-1] = tic_spacing[-1] - 1  # hack for edge effects
            plt.xticks(tic_spacing, tic_lables, fontsize=6)
            plt.yticks(tic_spacing, tic_lables, fontsize=6)
            # plt.xlabel('[um]', fontsize=8)
            # plt.ylabel('[um]', fontsize=8)

        # Z-axis scale
        if logZ:
            if np.min(slice) < 0:
                im = ax.imshow(slice, interpolation='none', origin='lower',
                               vmin=vlim[0], vmax=vlim[1],
                               norm=SymLogNorm(linthresh=1e-5),
                               cmap="YlGnBu_r")
                clabel = "Log Normalized Intensity"
            else:
                im = ax.imshow(slice, interpolation='none', origin='lower',
                               vmin=vlim[0], vmax=vlim[1], norm=LogNorm(),
                               cmap="YlGnBu_r")
                clabel = "Log Normalized Intensity"
        else:
            im = ax.imshow(slice,
                           interpolation='none', origin='lower', vmin=vlim[0], vmax=vlim[1], cmap="YlGnBu_r")
            clabel = "Normalized Intensity"

        if use_axis == 'anno':
            ax.annotate_axis(im, ax, datacube.shape[1])
        if use_axis is None:
            plt.axis('off')

    if use_axis:
        warnings.simplefilter("ignore", category=UserWarning)
        gs.tight_layout(fig, pad=1.08, rect=(0, 0, 1, 0.85))  # rect = (left, bottom, right, top)
        # fig.tight_layout(pad=50)
        cbar_ax = fig.add_axes([0.55, 0.3, 0.2, 0.05])  # Add axes for colorbar @ position [left,bottom,width,height]
        cb = fig.colorbar(im, cax=cbar_ax, orientation='horizontal')  #
        cb.set_label(clabel)

    if show is True:
        plt.show(block=True)


def view_timeseries(img_tseries, title=None, show=True, logZ=False, use_axis=True, vlim =(None,None),
                    dx=None, subplt_cols=3):
    """
    view white light images in the timeseries

    :param img_tseries: complex timeseries
    :param title: string, must be set or will error!
    :param show: flag possibly useful for plotting loops of things?
    :param logZ: turn logscale plotting for Z-axis on or off
    :param use_axis: turn on/off using axis ticks, colorbar, etc
    :param vlim: tuple of colorbar axis limits (min,max)
    :param subplt_cols: number of subplots per row
    :param dx: sampling of the image in m. Hardcoded to convert to um
    :return:
    """
    plt.close('all')

    # Recreate CDI phase stream for plot titles
    if cdi.use_cdi:
        phases = cdi.phase_series

    # Create figure & adjust subplot number, layout, size, whitespace
    n_tsteps = len(img_tseries)
    n_rows = int(np.ceil(n_tsteps / float(subplt_cols)))

    fig, subplot = plt.subplots(n_rows, subplt_cols, figsize=(10, 10))
    fig.subplots_adjust(bottom=0.1, top=0.85, hspace=.4)  #  wspace=0.2, right=0.95, left=0.05,

    # Title
    if title is None:
        warnings.warn("Plots without titles: Don't Do It!")
        title = input("Please Enter Title: ")
        pass
    fig.suptitle(title, fontweight='bold', fontsize=16)

    for ax, t in zip(subplot.flatten(), range(n_rows*subplt_cols)):  # range(n_tsteps)
        if t > n_tsteps-1:
            # ax.set_aspect('equal')
            ax.axis('off')  # hides axis
            pass
        else:
            # X,Y lables
            if dx is not None:
                # Converting Sampling Units to Readable numbers
                if dx < 1e-6:
                    dx *= 1e6  # [convert to um]
                    axlabel = 'um'
                elif dx < 1e-3:
                    dx *= 1e3  # [convert to mm]
                    axlabel = 'mm'
                elif 1e-2 > dx > 1e-3:
                    dx *= 1e2  # [convert to cm]
                    axlabel = 'cm'
                else:
                    axlabel = 'm'

                # Setting Tick Spacing
                tic_spacing = np.linspace(0, img_tseries[t].shape[0], 5)  # 5 (# of ticks) is just set by hand, arbitrarily chosen
                tic_lables = np.round(
                    np.linspace(-dx * sp.maskd_size / 2, dx * sp.maskd_size / 2, 5)).astype(
                    int)  # nsteps must be same as tic_spacing
                tic_spacing[0] = tic_spacing[0] + 1  # hack for edge effects
                tic_spacing[-1] = tic_spacing[-1] - 1  # hack for edge effects
                plt.xticks(tic_spacing, tic_lables, fontsize=6)
                plt.yticks(tic_spacing, tic_lables, fontsize=6)
                # plt.xlabel('[um]', fontsize=8)
                plt.ylabel(axlabel, fontsize=8)

            if logZ:
                if vlim[0] is not None and vlim[0] <= 0:
                    if cdi.use_cdi and not np.isnan(phases[t]):
                        ax.set_title(f"probe " r'$\theta$' + f"={phases[t]/np.pi:.2f}" + r'$\pi$')
                    else:
                        ax.set_title(f"t={t*sp.sample_time}")
                    im = ax.imshow(img_tseries[t], interpolation='none', origin='lower',
                                   vmin=vlim[0], vmax=vlim[1],
                                   norm=SymLogNorm(linthresh=1e-5), cmap="YlGnBu_r")
                    clabel = "Log Normalized Intensity"
                else:
                    if cdi.use_cdi and not np.isnan(phases[t]):
                        ax.set_title(f"probe" r'$\theta$' + f"={phases[t]/np.pi:.2f}" + r'$\pi$')
                    else:
                        ax.set_title(f"t={t * sp.sample_time}")
                    im = ax.imshow(img_tseries[t], interpolation='none', origin='lower',
                                   vmin=vlim[0], vmax=vlim[1],
                                   norm=LogNorm(), cmap="YlGnBu_r")
                    clabel = "Log Normalized Intensity"
            else:
                if cdi.use_cdi and not np.isnan(phases[t]):
                    ax.set_title(f"t={t * sp.sample_time},\nprobe" r'$\theta$' + f"={phases[t]/np.pi:.2f}" + r'$\pi$')
                else:
                    ax.set_title(f"t={t * sp.sample_time}\n")
                im = ax.imshow(img_tseries[t], interpolation='none', origin='lower',
                               vmin=vlim[0], vmax=vlim[1],
                               cmap="YlGnBu_r")
                clabel = "Normalized Intensity"

            if use_axis == 'anno':
                ax.annotate_axis(im, ax, img_tseries.shape[1])
            if use_axis is None:
                plt.axis('off')

        if use_axis:
            warnings.simplefilter("ignore", category=UserWarning)
            cbar_ax = fig.add_axes([0.85, 0.1, 0.05, 0.8])  # Add axes for colorbar @ position [left,bottom,width,height]
            cb = fig.colorbar(im, cax=cbar_ax, orientation='vertical')  #
            cb.set_label(clabel, fontsize=12)

    if show is True:
        plt.tight_layout(rect=[0, 0, 0.85, 0.9])  # rect = (left, bottom, right, top)
        plt.show(block=True)


def plot_planes(cpx_seq, title=None, logZ=[False], use_axis=True, vlim=[None, None], subplt_cols=3,
                 dx=None, first=False):
    """
    view plot of intensity in each wavelength bin at a single (last) timestep
    will pull out the plane(s) of sp.save_list at last tstep of cpx_sequence, convert to intensity, and sum over
      wavelength and object

    Currently, the atmosphere and enterance pupil are plotted in units of phase vs intensity. I think we should change
    this later for user-specification

    :param cpx_seq:
    :param title: string, must be set or will error!
    :param logZ: turn logscale plotting for z-axis on or off
    :param use_axis: turn on/off using axis ticks, colorbar, etc
    :param vlim: tuple of colorbar axis limits (min,max)
    :param subplt_cols: number of subplots per row
    :param dx: sampling of the image at each saved plane
    :return:
    """
    plt.close('all')

    # Create figure & adjust subplot number, layout, size, whitespace
    fig = plt.figure()
    n_planes = len(sp.save_list)
    n_rows = int(np.ceil(n_planes / float(subplt_cols)) )
    plt.axis('off')
    gs = gridspec.GridSpec(n_rows, subplt_cols, wspace=0.08)

    # Main Title
    if title is None:
        warnings.warn("Plots without titles: Don't Do It!")
        title = input("Please Enter Title: ")
        pass
    fig.suptitle(title, fontweight='bold', fontsize=16)

    # Small Hack to repeat axis if defaults used
    np.array(vlim)
    if len(logZ) == 1:
        logZ = np.repeat(logZ,len(sp.save_list))
    if len(vlim) == 2:
        vlim = (vlim,)*len(sp.save_list)

    for p in range(n_planes):
        ax = fig.add_subplot(gs[p])

        ###################
        # Retreiving Data
        ##################
        # Standard-Way
        # [timestep, plane, wavelength, object, x, y]
        # converts to intensity of last timestep, THEN sums over wavelength, then sums over object
        plot_plane = sp.save_list[p]
        plane = opx.extract_plane(cpx_seq, plot_plane)
        # Distinguish plotting z-axis in phase units or intensity units
        if plot_plane == "atmosphere" or plot_plane == "entrance_pupil":
            if first:
                plane = np.sum(np.angle(plane[0]), axis=(0,1))  # first timestep
            else:
                plane = np.sum(np.angle(plane[-1]), axis=(0,1))  # last timestep
            plane = opx.extract_center(plane, new_size=np.int(sp.grid_size*sp.beam_ratio)+10)
            logZ[p] = False
            vlim[p] = [None, None]
            phs = " phase"
        elif plot_plane == "woofer" or plot_plane == "tweeter" or plot_plane == "DM":
            # only show the star phase map since phase at other bodies just offsets to shift focal plane position
            if first:
                plane = np.sum(np.angle(plane[0]), axis=(0))  # first timestep, only sum over object
            else:
                plane = np.sum(np.angle(plane[-1]), axis=(0))  # last timestep, only sum over object
            plane = plane[0]  # plot the shortest wavelength
            plane = opx.extract_center(plane, new_size=np.int(sp.grid_size*sp.beam_ratio)+10)  # zoom in on DM
            logZ[p] = False
            vlim[p] = [-np.pi/100, np.pi/100]
            phs = " phase"
        else:
            if first:
                plane = np.sum(opx.cpx_to_intensity(plane[0]), axis=(0, 1))
            else:
                plane = np.sum(opx.cpx_to_intensity(plane[-1]), axis=(0,1))
            phs = ""
        ### Retreiving Data- Custom selection of plane ###
        # plot_plane = sp.save_list[w]
        # plane = opx.extract_plane(cpx_seq, plot_plane)
        # # plane = opx.cpx_to_intensity(plane[-1])
        # plane = opx.extract_center(np.angle(plane[0,1,1]))  # wavelengths, objects

        # Converting Sampling Units to Readable numbers
        if dx[p,0] < 1e-5:
            dx[p,:] *= 1e6  # [convert to um]
            axlabel = 'um'
        elif dx[p,0] < 1e-3:
            dx[p,:] *= 1e3  # [convert to mm]
            axlabel = 'mm'
        elif 1e-2 > dx[p,0] > 1e-3:
            dx[p,:] *= 1e2  # [convert to cm]
            axlabel = 'cm'
        else:
            axlabel = 'm'

        # X,Y lables
        if dx is not None:
            # dprint(f"sampling = {sampl[w]}")
            tic_spacing = np.linspace(0, plane.shape[0], 5)  # 5 (# of ticks) is just set by hand, arbitrarily chosen
            tic_lables = np.round(
                np.linspace(-dx[p,0] * plane.shape[0] / 2, dx[p,0] * plane.shape[0] / 2, 5)).astype(int)  # nsteps must be same as tic_spacing
            tic_spacing[0] = tic_spacing[0] + 1  # hack for edge effects
            tic_spacing[-1] = tic_spacing[-1] - 1  # hack for edge effects
            plt.xticks(tic_spacing, tic_lables, fontsize=6)
            plt.yticks(tic_spacing, tic_lables, fontsize=6)
            plt.ylabel(axlabel, fontsize=8)

        # Z-axis scale
        if phs == ' phase':
            cmap = sunlight
        else:
            cmap = "YlGnBu_r"
        if logZ[p]:
            if vlim[p][0] is not None and vlim[p][0] <= 0:
                ax.set_title(f"{sp.save_list[p]}"+phs)
                im = ax.imshow(plane, interpolation='none', origin='lower', vmin=vlim[p][0], vmax=vlim[p][1],
                               norm=SymLogNorm(linthresh=1e-5),
                               cmap=cmap)
                cb = fig.colorbar(im)
                # clabel = "Log Normalized Intensity"
            else:
                ax.set_title(f"{sp.save_list[p]}"+phs)
                im = ax.imshow(plane, interpolation='none', origin='lower', vmin=vlim[p][0], vmax=vlim[p][1],
                               norm=LogNorm(), cmap=cmap)  #(1e-6,1e-3)
                cb = fig.colorbar(im)
                # clabel = "Log Normalized Intensity"
                # cb.set_label(clabel)
        else:
            ax.set_title(f"{sp.save_list[p]}"+phs)
            im = ax.imshow(plane, interpolation='none', origin='lower', vmin=vlim[p][0], vmax=vlim[p][1],
                           cmap=cmap)  #  "twilight"
            cb = fig.colorbar(im)  #
            # clabel = "Normalized Intensity"
            # cb.set_label(clabel)

    if use_axis:
        warnings.simplefilter("ignore", category=UserWarning)
        gs.tight_layout(fig, pad=1.08, rect=(0, 0.02, 1, 0.9))  # rect = (left, bottom, right, top)
        # fig.tight_layout(pad=50)

    plt.show(block=True)



"""
model the SXExAO optical system starting with the DM and ending with the focal plane. This script is primarily used
to model the idealized focal plane response to probe patterns presented on the DM in the absence of other aberrations
or atmospheric effects on the telescope or optical train.

"""

import numpy as np
from inspect import getframeinfo, stack
import proper

from medis.params import sp, ap, tp
from medis.utils import dprint
import medis.optics as opx
import medis.aberrations as aber
import medis.adaptive as ao
import medis.coronagraphy as cg

#################################################################################################
#################################################################################################
#################################################################################################

# ------------------------------
# SCExAO
tp.entrance_d = 0.051  # diameter of optics in SCExAO train are 2 inches=0.051 m

tp.d_tweeter = 0.051  # diameter of optics in SCExAO train are 2 inches=0.051 m
tp.act_tweeter = 50  # approx a 2000 actuator DM, (45x45=2025)
tp.fl_SxOAPG = 0.255  # m focal length of Genera SCExAO lens (OAP1,3,4,5)
tp.fl_SxOAP2 = 0.519  # m focal length of SCExAO OAP 2
tp.d_SxOAPG = 0.051  # diameter of SCExAO OAP's

tp.dist_SxOAP1_scexao = 0.1345  # m
tp.dist_scexao_sl2 = 0.2511 - tp.dist_SxOAP1_scexao  # m
tp.dist_sl2_focus = 0.1261  # m

tp.lens_params = [{'aber_vals': [7.2e-17, 0.8, 3.1],
                   'diam': tp.d_SxOAPG,
                   'fl': tp.fl_SxOAPG,
                   'dist': tp.fl_SxOAPG,
                   'name': 'SxOAPG'},

                 {'aber_vals': [7.2e-17, 0.8, 3.1],
                   'diam': tp.d_SxOAPG,
                   'fl': tp.fl_SxOAP2,
                   'dist': tp.fl_SxOAP2,
                   'name': 'SxOAP2'}
                 ]
# ------------------------------
# Coronagraph
tp.cg_type = 'Gaussian'
# tp.cg_size = 3  # physical size or lambda/D size
tp.cg_size = 1.5  # physical size or lambda/D size
tp.cg_size_units = "l/D"  # "m" or "l/D"
tp.fl_cg_lens = tp.fl_SxOAPG  # m
tp.lyot_size = 0.95  # units are in fraction of surface un-blocked

#################################################################################################
#################################################################################################
#################################################################################################

def SCExAO_DM(empty_lamda, grid_size, PASSVALUE):
    """
    propagates instantaneous complex E-field thru Subaru from the primary through SCExAO

    this function is called a 'prescription' by proper

    uses PyPROPER3 to generate the complex E-field at the source, then propagates it through atmosphere,
        then telescope, to the focal plane
    the AO simulator happens here
    this does not include the observation of the wavefront by the detector
    :returns spectral cube at instantaneous time in the focal_plane()
    """
    # print("Propagating Broadband Wavefront Through Subaru")

    # Initialize the Wavefront in Proper
    wfo = opx.Wavefronts(sp.debug)
    wfo.initialize_proper()


    # Defines aperture (baffle-before primary)
    wfo.loop_collection(proper.prop_circular_aperture,
                        **{'radius': tp.entrance_d / 2})  # clear inside, dark outside
    wfo.loop_collection(proper.prop_define_entrance, plane_name='entrance_pupil')  # normalizes abs intensity

    # Test Sampling
    if sp.verbose:
        wfo.loop_collection(opx.check_sampling, PASSVALUE['iter'], "Telescope Aperture",
                            getframeinfo(stack()[0][0]), units='mm')
    # Testing Primary Focus (instead of propagating to focal plane)
    # wfo.loop_collection(opx.prop_pass_lens, tp.flen_nsmyth, tp.flen_nsmyth)  # test only going to prime focus

    ########################################
    # SCExAO
    # #######################################
    # AO System
    if tp.use_ao:
        WFS_map = ao.open_loop_wfs(wfo)
        wfo.loop_collection(ao.deformable_mirror, WFS_map, PASSVALUE['iter'], apodize=False,
                            plane_name='tweeter', debug=sp.verbose)
    # ------------------------------------------------
    wfo.loop_collection(proper.prop_propagate, tp.fl_SxOAPG)  # from tweeter-DM to OAP2

    # SXExAO Reimaging 2
    wfo.loop_collection(aber.add_aber, step=PASSVALUE['iter'], lens_name='SxOAP2')
    wfo.loop_collection(aber.add_zern_ab, tp.zernike_orders, aber.randomize_zern_values(tp.zernike_orders)/2)
    wfo.loop_collection(opx.prop_pass_lens, tp.fl_SxOAP2, tp.fl_SxOAP2, plane_name='post-DM-focus')  #tp.dist_sl2_focus

    # wfo.loop_collection(opx.check_sampling, PASSVALUE['iter'], "post-DM-focus",
    #                     getframeinfo(stack()[0][0]), units='nm')

    # Coronagraph
    # settings should be put into tp, and are not implicitly passed here
    wfo.loop_collection(cg.coronagraph, occulter_mode=tp.cg_type, plane_name='coronagraph')

    ########################################
    # Focal Plane
    # #######################################
    # Check Sampling in focal plane
    # wfo.focal_plane fft-shifts wfo from Fourier Space (origin==lower left corner) to object space (origin==center)
    cpx_planes, sampling = wfo.focal_plane()

    if sp.verbose:
        wfo.loop_collection(opx.check_sampling, PASSVALUE['iter'], "focal plane",
                            getframeinfo(stack()[0][0]), units='nm')
        # opx.check_sampling(PASSVALUE['iter'], wfo, "focal plane", getframeinfo(stack()[0][0]), units='arcsec')

    if sp.verbose:
        print(f"Finished datacube at timestep = {PASSVALUE['iter']}")

    return cpx_planes, sampling




import setuptools

setuptools.setup(
   name='medis',
   version='0.0.2',
   description='MKID Exoplanet Direct Imaging Simulator',
   author='R. Dodkins, K. Davis',
   author_email='dodkins@ucsb.edu',
   url="https://github.com/MazinLab/MEDIS",
   packages=setuptools.find_packages(),
)



"""
run_medis_main
Kristina Davis, Rupert Dodkins

This is the version of code compiled by K. Davis and R. Dodkins updated and streamlined from the MEDISv0.0 version of
the code written by R. Dodkins during his graduate career. This code contains the same functionality as MEDISv0.0 and is
modular to continue adding features and functionality.

MEDIS is split into two main functionalities; the first is to do an end-to-end simulation of a telescope. This is more
or less a large wrapper for proper with some new capabilities built into it to enable faster system design.The second
main functionality is to convert the complex field data generated by the telescope simulator to MKID-type photon lists.

For the telescope simulator, the user builds a prescription "name.py" of the optical system of the telescope system as
a separate module in the Simulations subdirectory of MEDIS. They can then update the default params file in a "
run_name.py" type script, which also makes the call to run_MEDIS to start the simulation. The output is a 6D complex
field observation sequence.

The MKIDs camera simulator creates an instance of a device with distributions in R, dead pixel locations etc and then
combines the fields sequence with this to generate realistic detected photons

"""
import os
import numpy as np
import pickle
from datetime import datetime
from pprint import pprint

from medis.telescope import Telescope
from medis.MKIDS import Camera
from medis.params import sp, ap, tp, iop, atmp, mp
from medis.CDI import cdi

################################################################################################################
################################################################################################################
################################################################################################################
sentinel = None


class RunMedis():
    """
    Creates a simulation for calling Telescope or MKIDs to return a series of complex electric fields or photons,
    respectively.

    This class is a wrapper for Telescope and Camera that handles the checking of testdir and params existence

    Upon creation the code checks if a testdir of this name already exists, if it does it then checks if the params
    match. If the params are identical and the desired products are not already created, if it will create them.
    If the params are different or the testdir does not already exist a new testdir and simulation is created.

    """
    def __init__(self, name='test', product='fields'):
        """
        File structure:
        datadir
            testdir          <--- output
                params.pkl   <--- output

        :param name:  used at the folder name where all the data is stored
        :param product:
            fields - 6D complex tensor (t,plane,wavelength,object,x,y) product of Telescope()
            photons - photon list with non-ideal MKID affects applied
            rebinned_cube - 4D tensor (t,wavelength,x,y) like scaled_fields but with non-ideal MKID affects aplied

        """

        self.name = name
        self.product = product
        assert product in ['fields', 'photons', 'rebinned_cube'], f"Requested data product {self.product} not supported"

        iop.update_testname(self.name)
        # for storing a checking between tests
        self.params = {'ap': ap, 'tp': tp, 'atmp': atmp, 'cp': cdi, 'iop': iop, 'sp': sp, 'mp': mp}

        # show all the parameters input into the simulation (before some are updated by Telescope and Camera)
        if sp.verbose:
            for param in self.params.values():
                print(f'\n\t {param.__name__()}')
                pprint(param.__dict__)

        # always make the top level directory if it doesn't exist yet
        if not os.path.isdir(iop.datadir):
            print(f"Top level directory... \n\n\t{iop.datadir} \n\ndoes not exist yet. Creating")
            os.makedirs(iop.datadir, exist_ok=True)

        if not os.path.exists(iop.testdir) or not os.path.exists(iop.params_logs):
            print(f"No simulation data found at... \n\n\t{iop.testdir} \n\n A new test simulation"
                  f" will be started")
            self.make_testdir()
        else:
            params_match = self.check_params()
            exact_match = all(params_match.values())
            if exact_match:
                print(f"Configuration files match. Initialization over")
            else:
                print(f"Configuration files differ")
                now = datetime.now().strftime("%m:%d:%Y_%H-%M-%S")
                backup_testr = os.path.join(iop.datadir, self.name+'_backup_'+now)
                if not sp.auto_load:
                    choice = input(f'\n\n\tINPUT REQUIRED...\n'
                                    f'Rename old testdir to {backup_testr} and start new simulation as {iop.testdir} [R],'
                                    f'\nor Quit [Q],'
                                    f'\nor overwrite the pickle file .pkl (keeps atmosphere and aberration maps) [P]?'
                                    f'\nor ignore parameter difference and proceed with Loading original [L]?\n\n')

                    if choice.lower() == 'q':
                        exit()
                    elif choice.lower() == 'r':
                        os.rename(iop.testdir, backup_testr)
                        self.make_testdir()
                    elif choice.lower() == 'p':
                        if os.path.exists(f"{iop.testdir}/fields.h5"):
                            # Don't overwrite the .pkl file if there is an existing h5 file.
                            print(f"Existing h5 file found. \n"
                                  f"Recommend saving as backup [R]\n"
                                  f"Exiting\n")
                            exit()
                        # Overwrite the .pkl files but keep aberration and atmosphere directories
                        os.remove(f"{iop.testdir}/params.pkl")

    def make_testdir(self):
        if not os.path.isdir(iop.testdir):
            os.makedirs(iop.testdir, exist_ok=True)

        with open(iop.params_logs, 'wb') as handle:
            pickle.dump(self.params, handle, protocol=pickle.HIGHEST_PROTOCOL)

    def check_params(self):
        """ Check all param classes at this stage. Some params are updated in Telescope and Camera and those params
         should be checked against each other at that stage of the pipeline"""
         #TODO check that the naming conventions of atmosphere and aberration maps match the new file format rather
         #TODO than only checking the directory name
        with open(iop.params_logs, 'rb') as handle:
            loaded_params = pickle.load(handle)

        match_params = {}
        print(f"\nChecking Matching Params Classes:")
        for p in ['ap','tp','atmp','iop','sp','mp','cp']:
            matches = []
            for (this_attr, this_val), (load_attr, load_val) in zip(self.params[p].__dict__.items(),
                                                                    loaded_params[p].__dict__.items()):
                try:
                    match = this_attr == load_attr and np.all(load_val == this_val)
                except ValueError:
                    match = False

                if match == False:
                    print(f'\n\tmismatch found: this_attr= {this_attr}, load_attr= {load_attr}, this_val= {this_val}, '
                          f'load_val= {load_val}')
                matches.append(match)

            match = np.all(matches)
            print(f"param: {p}, match: {match}")
            match_params[p] = match

        return match_params

    def __call__(self, *args, **kwargs):
        """ Get fields from Telescope and optionally then get photons from Camera. This looks complicated because of
         the possibility to chunk in both Telescope and Camera but simplifies a lot if both have num_chunk = 1"""

        if self.product == 'fields':
            # get the telescope configuration
            self.tel = Telescope(usesave=sp.save_to_disk)  # checking of class's cache etc is left to the class

            #load or generate the fields
            observation = self.tel()

        else:
            self.cam = Camera(usesave=sp.save_to_disk, product=self.product)  # get the camera configuratoin
            if self.cam.photontable_exists:
                observation = self.cam()
            else:
                self.tel = Telescope(usesave=sp.save_to_disk)
                self.tel()

                if self.tel.num_chunks == 1:
                    observation = self.cam(fields=self.tel.cpx_sequence)
                else:
                    for ichunk in range(int(self.tel.num_chunks)):
                        fields = self.tel.load_fields(span=(ichunk*self.tel.chunk_steps, (ichunk+1)*self.tel.chunk_steps))['fields']

                        observation = self.cam(fields=fields, abs_step=ichunk*self.tel.chunk_steps,
                                               finalise_photontable=False)

                    if self.product == 'photons' and not self.cam.photontable_exists:
                        self.cam.save_photontable(photonlist=[], index=('ultralight', 6), populate_subsidiaries=True)
                        self.cam.photontable_exists = True
                        self.cam.save_instance()

                    print('Returning the observation data for the final chunk only')

        return observation


if __name__ == '__main__':
    sim = RunMedis(name='example1', product='photons')
    observation = sim()
    print(observation.keys())





"""
optics.py

contains functionality related to the optical components needed to build up a proper prescription. This is a generic
module containing functions that don't have a home elsewhere. It contains the class structure to make the wavefront
object used in most prescriptions.
"""
import numpy as np
import proper
import copy
from scipy.interpolate import interp1d
from scipy.ndimage import gaussian_filter
from skimage.restoration import unwrap_phase
import matplotlib.pylab as plt
from matplotlib.colors import LogNorm
from inspect import getframeinfo, stack

from medis.twilight_colormaps import sunlight
from medis.params import ap, tp, sp
from medis.utils import dprint
from medis.distribution import planck


class Wavefront(proper.WaveFront):
    """ Wrapper for proper.Wavefront that stores source and wavelength info """

    def __init__(self, wavefront, lamda, name, beam_ratio, iw, ib):
        self.lamda = lamda
        self.name = name
        self.beam_ratio = beam_ratio
        self.iw = iw
        self.ib = ib

        for attr in dir(wavefront):
            if not hasattr(self, attr):
                setattr(self, attr, getattr(wavefront, attr))


class Wavefronts():
    """
    An object containing all of the complex E fields for each sampled wavelength and astronomical object at this tstep

    :params

    :returns
    self.wf_collection: a 2D array, where each element is its own 2D proper wavefront object structured as
     self.wf_collection[iw,ib]. Here, iw, ib is each wavefront and object, respectively
        ...meaning its an array of arrays.
        thus, self.wf_collection[iw,ib] is itself a 2D array of complex data. Its size is [sp.grid_size, sp.grid_size]
        we will call each instance of the collection a single wavefront wf
    self.save_E_fields: a matrix of E fields (proper.WaveFront.wfarr) at specified locations in the chain
    """
    def __init__(self, debug=False):

        # Using Proper to propagate wavefront from primary through optical system, loop over wavelength
        self.debug = debug
        self.wsamples = np.linspace(ap.wvl_range[0], ap.wvl_range[1], ap.n_wvl_init)  # units set in params (should be m)
        self.num_bodies = 1 + len(ap.contrast) if ap.companion else 1
        # wf_collection is an array of arrays; the wf_collection is (number_wavelengths x number_astro_bodies)
        # each 2D field in the wf_collection is the 2D array of complex E-field values at that wavelength, per object
        # the E-field size is given by (sp.grid_size x sp.grid_size)

        ############################
        # Create Wavefront Array
        ############################
        self.wf_collection = np.empty((len(self.wsamples), self.num_bodies), dtype=object)

        # Init Locations of saved E-field
        self.saved_planes = []  # string of locations where fields have been saved (should match sp.save_list after run is completed)
        self.Efield_planes = np.empty((0, np.shape(self.wf_collection)[0],  # array of saved complex field data at
                                           np.shape(self.wf_collection)[1],    # specified locations of the optical train
                                           sp.grid_size,
                                           sp.grid_size), dtype=np.complex64)
        # self.plane_sampling = np.empty((len(sp.save_list), ap.n_wvl_init))
        self.plane_sampling = []

        self.spectra = []
        for object_spectrum in ap.spectra:
            if isinstance(object_spectrum, (int, float)):
                spectrum = planck(object_spectrum, self.wsamples)
                spectrum /= np.sum(spectrum)
            else:
                spectrum = np.ones((ap.n_wvl_init))
            self.spectra.append(spectrum)

    def initialize_proper(self, set_up_beam=False):
        """

        Initialize the Wavefronts in Proper

        :param set_up_beam: bool applies prop_circular_aperture and prop_define_entrance before spectral scaling
            instead of during prescription wher it normally goes

        returns wf_colllection attribute array of wavefronts
        """
        for iw, wavelength in enumerate(self.wsamples):
            # Scale beam ratio by wavelength for polychromatic imaging
            # see Proper manual pg 37
            # Proper is devised such that you get a Nyquist sampled image in the focal plane. If you optical system
            #  goes directly from pupil plane to focal plane, then you need to scale the beam ratio such that sampling
            #  in the focal plane is constant. You can check this with check_sampling, which returns the value from
            #  prop_get_sampling. If the optical system does not go directly from pupil-to-object plane at each optical
            #  plane, the beam ratio does not need to be scaled by wavelength, because of some optics wizardry that
            #  I don't fully understand. KD 2019
            if sp.focused_sys:
                beam_ratio = sp.beam_ratio
            else:
                beam_ratio = sp.beam_ratio * ap.wvl_range[0] / wavelength
                # dprint(f"iw={iw}, w={w}, beam ratio is {self.beam_ratios[iw]}")

            # Initialize the wavefront at entrance pupil
            wfp = proper.prop_begin(tp.entrance_d, wavelength, sp.grid_size, beam_ratio)
            if set_up_beam:
                proper.prop_circular_aperture(wfp, radius = tp.entrance_d / 2)
                proper.prop_define_entrance(wfp)  # normalizes the intensity
            wfp.wfarr = np.multiply(wfp.wfarr, np.sqrt(self.spectra[0][iw]), out=wfp.wfarr, casting='unsafe')

            wfs = [wfp]
            names = ['star']

            # Initiate wavefronts for companion(s)
            if ap.companion:
                for ix in range(len(ap.contrast)):
                    wfc = proper.prop_begin(tp.entrance_d, wavelength, sp.grid_size, beam_ratio)
                    if set_up_beam:
                        proper.prop_circular_aperture(wfc, radius=tp.entrance_d / 2)
                        proper.prop_define_entrance(wfc)  # normalizes the intensity
                    wfc.wfarr = np.multiply(wfc.wfarr, np.sqrt(self.spectra[ix][iw]), out=wfc.wfarr, casting='unsafe')
                    wfs.append(wfc)
                    names.append('companion_%i' % ix)

            for io, (name, wf) in enumerate(zip(names, wfs)):
                self.wf_collection[iw, io] = Wavefront(wf, wavelength, name, beam_ratio, iw, io)

    def loop_collection(self, func, *args, **kwargs):
        """
        For each wavelength and astronomical object apply a function to the wavefront.

        The wavefront object has dimensions of shape=(n_wavelengths, n_astro_bodies, grid_sz, grid_sz)

        To save, you must pass in the keyword argument plane_name when you call this function from the prescription.
        Specifically for the AO/DM system, there is different behavior in the function depending on the plane name.
        Therefore, we need to pass the plane_name through to ao.deformable_mirror. Otherwise, we just use it for saving
        and we can pop it out of the kwargs list because no other functions use it for dependant functionality.

        If you are saving the plane at this location, keep in mind it is saved AFTER the function is applied. This
        is desirable for most functions but be careful when using it for prop_lens, etc

        :param func: function to be applied e.g. ap.add_aber()
        :param args: args to be passed to the function
        :param kwargs: kwargs to be passed to the function
        :return: everything is just applied to the wfo, so nothing is returned in the traditional sense
        """
        if 'plane_name' in kwargs:
            if kwargs['plane_name'] == 'woofer' or kwargs['plane_name'] == 'tweeter':
                plane_name = kwargs['plane_name']
                pass  # keep the plane_name in kwargs.
            else:
                plane_name = kwargs.pop('plane_name')  # remove plane_name from **kwargs
        elif 'lens_name' in kwargs:
            plane_name = kwargs['lens_name']
        else:
            plane_name = func.__name__

        if 'zero_outside' in kwargs:
            raise DeprecationWarning('Zero outside issues fixed by pupil masking in adaptive.py')

        # manipulator_output is a way to store the values of a function passed into loop_collections.
        # EG you could use this as a way to save your WFS map if desired. The WFS map would be returned
        # as a 2D array by a WFS function (eg ao.closed_loop_wfs()) and then we could call it later and use it
        # or  just save it and plot it
        # also if we don't use it we can get rid of some of it
        # manipulator_output = np.empty(self.wf_collection.shape)
        manipulator_output = [[[] for _ in range(len(self.wsamples))] for _ in range(self.num_bodies)]
        for iw, sources in enumerate(self.wf_collection):
            for io, wavefront in enumerate(sources):
                if not func.__name__ in sp.skip_functions:
                    manipulator_output[io][iw] = func(wavefront, *args, **kwargs)

        # Show phase and amplitude of the plane during debugging
        if self.debug and not func.__name__ in sp.skip_functions:
            self.quicklook(title=plane_name)

        manipulator_output = np.array(manipulator_output)
        # if there's at least one not np.nan element add this array to the Wavefronts obj
        if np.any(manipulator_output != None):
            setattr(self, plane_name, manipulator_output)

        # Saving complex field data after function is applied
        if plane_name in sp.save_list:
            self.save_plane(location=plane_name)

    def save_plane(self, location=None):
        """
        Saves the complex field at a specified location in the optical system. If the function is called by
        wfo.loop_collection, the plane is saved AFTER the function is applied

        Note that the complex planes saved are not summed by object, interpolated over wavelength, nor masked down
        to the sp.maskd_size.

        :param location: name of plane where field is being saved
        :return: self.save_E_fields
        """
        if sp.verbose:
            dprint(f"saving plane at {location}")

        if location is not None and location in sp.save_list:
            E_field = np.zeros((1, np.shape(self.wf_collection)[0],
                                       np.shape(self.wf_collection)[1],
                                       sp.grid_size,
                                       sp.grid_size), dtype=np.complex64)
            samp_lambda = np.zeros(ap.n_wvl_init)

            for iw, sources in enumerate(self.wf_collection):
                samp_lambda[iw] = proper.prop_get_sampling(self.wf_collection[iw,0])
                for io, wavefront in enumerate(sources):
                    wf = proper.prop_shift_center(wavefront.wfarr)
                    E_field[0, iw, io] = copy.copy(wf)

            self.Efield_planes = np.vstack((self.Efield_planes, E_field))
            self.saved_planes.append(location)
            self.plane_sampling.append(samp_lambda)

    def focal_plane(self):
        """
        ends the proper prescription and return sampling. most functionality involving image processing now in utils

        :return:
        """
        # Saving Complex Data via save_plane
        self.save_plane(location='detector')           # shifting, etc already done in save_plane function

        cpx_planes = np.array(self.Efield_planes)
        sampling = np.array(self.plane_sampling)

        # Conex Mirror-- cirshift array for off-axis observing
        # if tp.pix_shift is not [0, 0]:
        #     datacube = np.roll(np.roll(datacube, tp.pix_shift[0], 1), tp.pix_shift[1], 2)

        return cpx_planes, sampling

    def quicklook(self, wf=None, logZ=True, show=True, title=None):
        """
        Produces a figure with an image of amplitude and one of phase as well as 1D slices through these images

        :param wf: optics.Wavefront
        :param logZ: bool logarithmic Z scaling
        :param show: bool display figure now or leave show() to be called by user later on
        :param title: str
        :return:
        """
        if wf == None:
            wf = self.wf_collection[0,0]

        amp_map = proper.prop_get_amplitude(wf)
        phase_map = proper.prop_get_phase(wf)

        fig = plt.figure(figsize=(12, 10))
        ax1 = plt.subplot2grid((1, 2), (0, 0), rowspan=2)  # ((shape of grid to place axis x,y),(location x,y))
        ax2 = plt.subplot2grid((1, 2), (0, 1), rowspan=2)
        # ax3 = plt.subplot2grid((3, 2), (2, 0))
        # ax4 = plt.subplot2grid((3, 2), (2, 1))
        if logZ:
            im1 = ax1.imshow(amp_map, origin='lower', cmap="YlGnBu_r", norm=LogNorm())
        else:
            im1 = ax1.imshow(amp_map, origin='lower', cmap="YlGnBu_r")
        ax1.set_title('Amplitude Map')
        plt.colorbar(im1, ax=ax1)

        im2 = ax2.imshow(phase_map, origin='lower', cmap=sunlight)  # , vmin=-0.5, vmax=0.5)
        ax2.set_title('Phase Map')
        plt.colorbar(im2, ax=ax2)

        # ax3.plot(after_dm[int(sp.grid_size / 2)])
        # ax3.plot(np.sum(np.eye(sp.grid_size) * after_dm, axis=1), label=f'row {int(sp.grid_size / 2)}')
        # ax3.legend()
        #
        # # plt.plot(np.sum(after_dm,axis=1)/after_dm[128,128])
        #
        # ax4.plot(phase_afterdm[int(sp.grid_size / 2)], label=f'row {int(sp.grid_size / 2)}')
        # ax4.legend()
        # # ax4.plot(np.sum(np.eye(ap.grid_size)*phase_afterdm,axis=1))
        # plt.xlim([0, proper.prop_get_gridsize(wf)])
        if not title:
            title = input(f"Always Add A Title\n "
                          f"Please Enter Plane Name:")
            fig.suptitle(f"plane: {title}, lambda: {wf.lamda} m, body: {wf.name}", fontsize=18)
        else:
            fig.suptitle(f"plane: {title}, lambda: {wf.lamda} m, body: {wf.name}", fontsize=18)

        plt.subplots_adjust(top=0.9)

        if show:
            plt.show(block=True)


####################################################################################################
# Functions Relating to Processing Complex Cubes
####################################################################################################
def interp_wavelength(data_in, ax):
    """
    Interpolating spectral cube from ap.n_wvl_init discreet wavelengths to ap.n_wvl_final

    :param data_in array where one axis contains the wavelength of the data
    :param ax  axis of wavelength
    :return data_out array that has been interpolated over axis=ax
    """
    # Interpolating spectral cube from ap.n_wvl_init discreet wavelengths to ap.n_wvl_final
    if ap.interp_wvl and 1 < ap.n_wvl_init < ap.n_wvl_final:
        wave_samps = np.linspace(0, 1, ap.n_wvl_init)
        f_out = interp1d(wave_samps, data_in, axis=ax)
        new_heights = np.linspace(0, 1, ap.n_wvl_final)
        data_out = f_out(new_heights)
    else:
        data_out = data_in

    return data_out


def interp_sampling(sampling):
    if ap.interp_wvl and 1 < ap.n_wvl_init < ap.n_wvl_final:
        wave_samps = np.linspace(0, 1, ap.n_wvl_init)
        f_out = interp1d(wave_samps, sampling, axis=1)  # axis 1 because sa
        new_heights = np.linspace(0, 1, ap.n_wvl_final)
        data_out = f_out(new_heights)
    else:
        data_out = sampling

    return data_out


def extract_plane(data_in, plane_name):
    """
    pull out the specified plane of the detector from complex array

    here we assume that the data_in has the specific format of:
    [timestep, plane, object, wavelength, x, y]
      meaning that no dimensions have been removed from the original obs sequence
    Code will return invalid results if data_in is not in this format

    :param data_in: the 5D wavefront array of shape  [timestep, plane, wavelength, x, y]
    :param plane_name: the name of a plane you want to pull out, must match the plane name given in sp.plane_list

    :return sequence of data with format [tstep, obj, wavelength, x, y] (remove plane dimension)
    """
    ip = sp.save_list.index(plane_name)
    return data_in[:, ip, :, :, :, :]  # [tsteps, #wavelengths, x, y]--it automatically squeezes the plane axis


def cpx_to_intensity(data_in):
    """
    converts complex data to units of intensity

    WARNING: if you sum the data sequence over object or wavelength with simple case of np.sum(), must be done AFTER
    converting to intensity, else results are invalid
    """
    return np.abs(data_in)**2


def extract_center(wf, new_size=sp.maskd_size):
    """
    extracts [new_size, new_size] from [sp.grid_size, sp.grid_size] data
    fp~focal plane
    code modified from the EXTRACT flag in prop_end

    :param wf: [sp.grid_size, sp.grid_size] array
    :returns: array with size [new_size, new_size]
    """
    smaller_wf = np.zeros((new_size, new_size))
    EXTRACT = new_size
    nx,ny = wf.shape
    smaller_wf = wf[int(ny/2-EXTRACT/2):int(ny/2+EXTRACT/2),
                    int(nx/2-EXTRACT/2):int(nx/2+EXTRACT/2)]
    return smaller_wf


####################################################################################################
# Functions Relating to Masking the Pupil Plane
####################################################################################################
def circular_mask(h, w, center=None, radius=None):
    if center is None: # use the middle of the image
        center = (int(w/2), int(h/2))
    if radius is None: # use the smallest distance between the center and image walls
        radius = min(center[0], center[1], w-center[0], h-center[1])

    Y, X = np.ogrid[:h, :w]
    # Y, X = np.mgrid[:h, :w]
    dist_from_center = np.sqrt((X - center[0])**2 + (Y-center[1])**2)

    mask = dist_from_center <= radius

    return mask


def apodize_pupil(wf):
    phase_map = proper.prop_get_phase(wf)
    amp_map = proper.prop_get_amplitude(wf)
    h, w = wf.wfarr.shape[:2]
    wavelength = wf.lamda
    scale = ap.wvl_range[0] / wavelength
    inds = circular_mask(h, w, radius=scale * 0.95 * h * sp.beam_ratio / 2)  #TO DO remove hardcoded sizing, especially if use is default
    mask = np.zeros_like(phase_map)
    mask[inds] = 1
    smooth_mask = gaussian_filter(mask, 1.5, mode='nearest')  #TO DO remove hardcoded sizing, especially if use is default
    smoothed = phase_map * smooth_mask
    wf.wfarr = proper.prop_shift_center(amp_map * np.cos(smoothed) + 1j * amp_map * np.sin(smoothed))


def abs_zeros(wf):
    """
    zeros everything outside the pupil

    This function attempts to figure out where the edges of the pupil are by determining if either the real
     or imaginary part of the complex-valued E-field is zero at a gridpoint. If so, it sets the full cpmplex
     value to zero, so 0+0j
    """
    bad_locs = np.logical_or(np.real(wf) == -0, np.imag(wf) == -0)
    wf[bad_locs] = 0 + 0j

    return wf


################################################################################################################
# Optics in Proper
################################################################################################################
def add_obscurations(wf, M2_frac=0, d_primary=0, d_secondary=0, legs_frac=0.05, plane_name=None):
    """
    adds central obscuration (secondary shadow) and/or spider legs as spatial mask to the wavefront

    :param wf: 2D proper wavefront
    :param M2_frac: ratio of tp.entrance_d the M2 occupies
    :param d_primary: diameter of the primary mirror
    :param d_secondary: diameter of the secondary mirror
    :param legs_frac: fractional size of spider legs relative to d_primary
    :return: acts upon wfo, applies a spatial mask of s=circular secondary obscuration and possibly spider legs
    """
    if tp.obscure is False:
        pass
    else:
        # dprint('Including Obscurations')
        if M2_frac > 0 and d_primary > 0:
            proper.prop_circular_obscuration(wf, M2_frac * d_primary)
        elif d_secondary > 0:
            proper.prop_circular_obscuration(wf, d_secondary)
        else:
            raise ValueError('must either specify M2_frac and d_primary or d_secondary')
        if legs_frac > 0:
            proper.prop_rectangular_obscuration(wf, legs_frac, 1.3, ROTATION=-20, NORM=True)
            proper.prop_rectangular_obscuration(wf, 1.3, legs_frac, ROTATION=-20, NORM=True)


def prop_pass_lens(wf, fl_lens, dist):
    """
    pass the wavefront through a lens then propagate to the next surface

    :param wf: single wavefront of shape=(sp.grid_sz, sp.grid_sz)
    :param fl_lens: focal length in m
    :param dist: distance in m
    """
    proper.prop_lens(wf, fl_lens)
    proper.prop_propagate(wf, dist)


def offset_companion(wf):
    """
    offsets the companion wavefront using the 2nd and 3rd order Zernike Polynomials (X,Y tilt)
    companion(s) contrast and location(s) set in params

    We don't call this function via wfo.loop_collection because we need to know which object (io) we are on, which
    is not supported in the current format. This is the only acception to applying loop_collection

    Important: this function must be called AFTER any calls to proper.prop_define_entrance, which normalizes the
    intensity, because we scale the intensity of the planet relative to the star via the user-parameter ap.contrast.

    If you have a focused system, and do not scale the grid sampling of the system by wavelength, we account
    for that here (thus the if/else statements). This is because we shift the companion's location in the focal plane
    by proper.prop_zernikes, which scales the x/y tilt (zernike orders 2 and 3) by wavelength to account for the
    presumed resampling based on wavelength. We thus counteract that in the case of sp.focused_sys=True

    Wavelength/contrast scaling scales the contrast ratio between the star and planet as a function of wavelength.
    This ratio is given by ap.C_spec, and the scale ranges from 1/ap.C_spec to 1 as a function of ap.n_wvl_init. The
        gradient ap.C_spec should be chosen carefully to consider the number of wavelengths and spectral type of the
        star and planet in the simulation.

    :param wf: singe wavefront object from wfo.wf_collection, shape=(grid_sz, grid_sz)
    :return: nothing implicitly returned but the given wfo initiated in Wavefronts class has been altered to give the
        appropriate wavefront for a planet in the focal plane
    """
    if ap.companion is True and wf.name != 'star':
        cont_scaling = np.linspace(1. / ap.C_spec, 1, ap.n_wvl_init)

        # Shifting the Array
        if sp.focused_sys:
            # Scaling into lambda/D AND scaling by wavelength
            xloc = ap.companion_xy[wf.ib-1][0] * wf.lamda / tp.entrance_d \
                   * ap.wvl_range[0] / wf.lamda # * (-1)**(iw%2)
            yloc = ap.companion_xy[wf.ib-1][1] * wf.lamda / tp.entrance_d \
                    *  ap.wvl_range[0] / wf.lamda  # / (2*np.pi)   * (-1)**(iw%2)
        else:
            # Scaling Happens Naturally!
            xloc = ap.companion_xy[wf.ib-1][0]*1e-6
            yloc = ap.companion_xy[wf.ib-1][1]*1e-6
        proper.prop_zernikes(wf, [2, 3], np.array([xloc, yloc]))  # zernike[2,3] = x,y tilt

        ##############################################
        # Wavelength/Contrast  Scaling the Companion
        ##############################################
        wf.wfarr *= np.sqrt(ap.contrast[wf.ib-1])

        #TODO implement wavelength-dependant scaling
        # Wavelength-dependent scaling by cont_scaling
        # wf = wf * np.sqrt(ap.contrast[wf.ib-1] * cont_scaling[wf.iw])


####################################################################################################
# Check Sampling & Misc Tools
####################################################################################################
def check_sampling(wf, tstep, location, line_info, units=None):
    """
    checks the sampling of the wavefront at the given location and prints to console

    mostly reformats the output of proper.prop_get_sampling into human-readable format

    :param wf: wavefront object
    :param tstep: timestep, will only check for first timestep, so when tstep==0
    :param location: string that identifies where call is being made
    :param line_info: info on the line number and function name from where check_sampling was called
        example: getframeinfo(stack()[0][0])
        via: from inspect import getframeinfo, stack
    :param units: desired units of returned print statement; options are 'mm,'um','nm','arcsec','rad'
    :return: prints sampling to the command line
    """
    if tstep == 0 and wf.ib == 0:
        if wf.iw == 0:
            print(f"\nFrom {line_info.filename}:{line_info.lineno}\n Sampling at {location}")

        check_sampling = proper.prop_get_sampling(wf)
        if units == 'mm':
            print(f"sampling at wavelength={wf.lamda * 1e9:.0f}nm is {check_sampling*1e3:.4f} mm")
        elif units == 'um':
            print(f"sampling at wavelength={wf.lamda * 1e9:.0f}nm is {check_sampling*1e6:.1f} um")
        elif units == 'nm':
            print(f"sampling at wavelength={wf.lamda * 1e9:.0f}nm is {check_sampling*1e9:.1f} nm")
        elif units == 'arcsec':
            check_sampling = proper.prop_get_sampling_arcsec(wf)
            print(f"sampling at wavelength={wf.lamda * 1e9:.0f}nm is {check_sampling*1e3:.2f} mas")
        elif units == 'rad':
            check_sampling = proper.prop_get_sampling_radians(wf)
            print(f"sampling at wavelength={wf.lamda * 1e9:.0f}nm is {check_sampling:.3f} rad")
        else:
            print(f"sampling at wavelength={wf.lamda * 1e9:.0f}nm is {check_sampling} m")


def unwrap_phase_zeros(phasemap):
    """ combination of abs_zeros and masking allows phase unwrap to work without discontiuities sometimes occur """
    masked_phase = np.ma.masked_equal(phasemap, 0)
    unwrap = unwrap_phase(masked_phase, wrap_around=[False, False])
    unwrap[phasemap == 0] = 0
    return unwrap
